from pathlib import Path
import subprocess

bitmap_module_path = Path(__file__).parent.parent.parent / 'data_structures' / 'Bitmap' / '_bitmap'


def main():
    ps = subprocess.run([
        'python.exe',
        str(bitmap_module_path / 'setup.py'),
        'build',
    ], cwd=str(bitmap_module_path.parent), check=True)



from abc import ABC, abstractmethod
from typing import Iterable, Any, Tuple

from ...hycore import NeoIo


class ChangeEvent:
    def __init__(self, key, value):
        self.key = key
        self.value = value


class BackendABC(ABC):
    serializer = None
    support_types = None

    def __init_subclass__(cls, **kwargs):
        if cls.serializer is None:
            raise TypeError("BackendABC subclass must have a serializer attribute.")
        if cls.support_types is None:
            raise TypeError("BackendABC subclass must have a support_types attribute.")

    def __init__(self):
        self.file = None
        self.existing = False

        self._data = {}
        self._fd = NeoIo()
        self._fd.create = True

    def _support_1(self, type):
        return type in self.support_types

    def _support_2(self, type):
        for i in self.support_types:
            if issubclass(type, i):
                return True
        return False

    def support(self, type):
        return self._support_1(type) or self._support_2(type)

    def init(self, **kwargs):
        self._data = kwargs

    def get(self, key):
        return self._data.get(key)

    def set(self, key, value):
        self._data[key] = value
        self.on_change(ChangeEvent(key, value))

    def keys(self) -> Iterable[str]:
        return self._data.keys()

    def values(self) -> Iterable[Any]:
        return self._data.values()

    def items(self) -> Iterable[Tuple[str, Any]]:
        return self._data.items()

    def on_change(self, event: ChangeEvent):
        ...

    @abstractmethod
    def save(self):
        ...

    @abstractmethod
    def load(self):
        ...



from .auto import AutoRegDict


class MultiSet:
    def __init__(self, iterable=None):
        self._elements = set()
        self._data = AutoRegDict()
        self._data.default_value = 0
        if iterable is not None:
            for item in iterable:
                self.add(item)

    def add(self, item):
        self._elements.add(item)
        self._data[item] += 1

    def remove(self, item):
        self._data[item] -= 1
        if self._data[item] <= 0:
            del self._data[item]
            self._elements.remove(item)

    def clear(self):
        self._elements.clear()
        self._data.clear()

    def set(self):
        return self._elements

    def __contains__(self, item):
        return item in self._elements

    def __len__(self):
        return len(self._elements)

    def __iter__(self):
        return iter(self._elements)



from typing import *


@runtime_checkable
class Task(Protocol):
    start: int
    end: int
    completed_size: int
    start_time: float
    connect_time: float


@runtime_checkable
class Policy(Protocol):
    chunk_min_size: int
    max_chunk_count: int


@runtime_checkable
class CallbackMetaclass(Protocol):
    def create_task(self, start: int, end: int) -> Task: ...

    def get_total_size(self) -> int: ...

    def get_assign_policy(self) -> Policy: ...



class Template:
    def __init__(self, **tem):
        self._template = tem

    @property
    def template(self):
        return self._template.copy()


def match(dic, tem):
    from . import Dict as dictx
    if dictx.sub(tem, dic) and dictx.parent(tem, dic):
        return True
    else:
        return False


def sort(dic, tem):
    # 修改字典使字典符合模板
    for i in tem.template:
        if i not in dic:
            dic[i] = tem.template[i]
    return dic


def sub(template: Template, data: dict):
    if data is None:
        return False
    from . import Dict as dictx
    return dictx.sub(template.template, data)



from abc import abstractmethod

from .auto import AutoState, AutoStr


class NetPackage(AutoState, AutoStr):
    """
    使用时，子类的`_state_attrs` 和 `_str_attrs` 应显式指定，否则可能无法正常序列化
    """
    @abstractmethod
    def get(self): ...

    @classmethod
    def is_package(cls, package_object):
        return isinstance(package_object, cls) or issubclass(package_object.__class__, cls)


class Request(NetPackage):
    __state_attrs__ = ("header", "data")

    def __init__(self, header=None, data=None):
        self.header = {} if header is None else header
        self.data = data

    def get(self):
        return self.header, self.data


class Answer(NetPackage):
    __state_attrs__ = ("header", "result", "status")

    def __init__(self, header=None, result=None, status=None):
        self.header = header
        self.result = result
        self.status = status

    def get(self):
        return self.header, self.result, self.status


class Error(NetPackage):
    __state_attrs__ = ("header", "error", "reason")

    def __init__(self, header=None, error=None, reason=None):
        self.header = header
        self.error = error
        self.reason = reason

    def get(self):
        return self.header, self.error, self.reason


class Info(NetPackage):
    __state_attrs__ = ("header", "info")

    def __init__(self, header=None, info=None):
        self.header = header
        self.info = info

    def get(self):
        return self.header, self.info


class Action(NetPackage):
    __state_attrs__ = ("header", "action")

    def __init__(self, header=None, action=None):
        self.header = header
        self.action = action

    def get(self):
        return self.header, self.action



from typing import Any

from . import Serializer


class Combination(Serializer):
    """
    Combination of multiple serializers
    """
    def __init__(self, *args: Serializer):
        self._serializers = list(args)

    def append(self, serializer: Serializer):
        self._serializers.append(serializer)

    def extend(self, serializers: list[Serializer]):
        self._serializers.extend(serializers)

    def clear(self):
        self._serializers.clear()

    def set(self, serializers: list[Serializer]):
        self._serializers = serializers

    def index_at(self, __i):
        return self._serializers[__i]

    def pop(self, __i):
        return self._serializers.pop(__i)

    def dumps(self, data) -> bytes:
        for serializer in self._serializers:
            data = serializer.dumps(data)
        return data

    def loads(self, data) -> Any:
        for serializer in reversed(self._serializers):
            data = serializer.loads(data)
        return data



from typing import Sequence

from ..abc.types import ConfigType


class IntType(ConfigType, types=int):
    def load(self, value):
        self.set(int(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, int)


class StringType(ConfigType, types=str):
    def load(self, value):
        self.set(str(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, str)


class FloatType(ConfigType, types=float):
    def load(self, value):
        self.set(float(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, float),


class BooleanType(ConfigType, types=bool):
    def load(self, value):
        self.set(bool(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, bool)


class ListType(ConfigType, types=list):
    def load(self, value):
        self.set(list(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, Sequence)


class TupleType(ConfigType, types=tuple):
    def load(self, value):
        self.set(tuple(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, tuple)


class DictType(ConfigType, types=dict):
    def load(self, value):
        self.set(dict(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, dict)


class SetType(ConfigType, types=set):
    def load(self, value):
        self.set(set(value))

    @classmethod
    def validate(cls, value):
        return isinstance(value, set)


class BytesType(ConfigType, types=bytes):
    def load(self, value):
        self.set(bytes(value))

    def transform(self):
        if self.backend.support(bytes):
            return self.value
        else:
            return list(self.value)

    @classmethod
    def validate(cls, value):
        return isinstance(value, bytes)



def get_subclasses(cls):
    """
    获取所有子类
    """
    if cls is type:
        return []
    return cls.__subclasses__()


def get_subclass_counts(cls):
    """
    获取所有子类的数量
    """
    if cls is type:
        return 0
    return len(cls.__subclasses__())


def get_subclasses_recursion(cls):
    """
    递归地获取所有子类
    """
    if cls is type:
        return []
    return (
            cls.__subclasses__() + [g for s in cls.__subclasses__() for g in get_subclasses_recursion(s)]
    )


def get_subclass_counts_recursion(cls):
    """
    获取所有子类的数量
    """
    if cls is type:
        return 0
    return len(cls.__subclasses__()) + sum(get_subclass_counts_recursion(s) for s in cls.__subclasses__())



from __future__ import annotations
import builtins
from typing import Any, Protocol, runtime_checkable

from ..abc.types import ConfigType
from ..._hycore.utils import InstanceDict


@runtime_checkable
class IDict_Item(Protocol):
    value: ConfigItemInstance
    ins: Any


class ConfigItemInstance:
    def __init__(self, type, attr, key, default, parent: 'ConfigItem' = None):
        self.key, self.attr = key, attr
        self.type = type
        self.default = default
        self.instance = type(default)

        self.parent = parent

    def sync(self):
        self.key, self.attr, self.type, self.default = (
            self.parent.key, self.parent.attr, self.parent.type, self.parent.default)

    def set(self, value):
        self.instance.set(value)

    def get(self):
        return self.instance.get()

    def validate(self, value, error=False):
        res = self.type.validate(value)
        if not res and error:
            raise TypeError(f"{type(value)} is not a valid type")
        return res

    @property
    def value(self):
        return self.get()


class ConfigItem:
    def __init__(self, key, *, type: type[ConfigType], default: Any):
        self.type = type
        self.default = default

        self.attr = None
        self.key = key

        self.validate(default)

        if not isinstance(type, builtins.type):
            raise TypeError("type must be a ItemType")

        self._instances = InstanceDict()

    def validate(self, value):
        if not self.type.validate(value):
            raise TypeError(f"{type(value)} is not a valid type")

    def _instance(self, ins) -> IDict_Item:
        if ins not in self._instances:
            self._instances[ins] = ConfigItemInstance(self.type, self.attr, self.key, self.default, self)
        return self._instances[ins]

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self._instance(instance).value.get()

    def __set__(self, instance, value):
        item = self._instance(instance).value  # idict_item.value

        old, new = item.value, value

        item.set(value)

        instance.on_change(self.attr, old, new)

    def from_instance(self, instance) -> 'ConfigItemInstance':
        item_instance = self._instance(instance).value
        return item_instance



from time import sleep, time


class Clock:
    """
    时钟类，用于控制任务刷新间隔。
    如果任务消耗的时间大于刷新间隔时间，则sleep一段时间，以保持任务刷新间隔。
    如果任务消耗的时间小于刷新间隔时间，则sleep剩余一段时间，以保持任务刷新间隔。
    示例:
    ```python
    import random
    import time
    def random_sleep(a, b):
        sleep_time = random.randint(a, b)
        time.sleep(sleep_time)

    clock = Clock()
    while True:
        random_sleep(100, 2000)
        clock.strike(1.5)
    ```
    程序将保证每次任务执行时间为1.5秒，当任务消耗时间大于1.5秒时，程序将调整暂停时间，以保持任务刷新间隔。
    """
    def __init__(self):
        self.last_time = None

    def strike(self, interval):
        """
        确保任务以指定的间隔时间执行。
        :param interval: 刷新间隔时间，单位为秒
        """
        if self.last_time is None:
            self.last_time = time()

        current_time = time()
        elapsed_time = current_time - self.last_time
        if elapsed_time >= interval:
            sleep(interval - elapsed_time % interval)
        else:
            sleep(interval - elapsed_time)
        self.last_time = time()



import os
from typing import Union

from .type_func import get_type_name

origin_environ = None

ENV_ORIGIN = 0x1
ENV_NEW = 0x2


def init():
    """
    这个操作将会完全影响模块的运作初始值
    """
    global origin_environ
    origin_environ = os.environ.copy()
    return Environment()


def _split_env_string(string):
    if os.name == 'nt':
        value = string.split(';')
    elif os.name == 'posix':
        value = string.split(':')
    elif os.name == 'java':
        value = string.split(':')
    else:
        raise NotImplementedError(f"{os.name} is not supported")
    return value


class EnvironmentItem:
    __slots__ = ('name', 'value')

    def __init__(self, name, value):
        # print(name)
        self.name = name
        self.value = value

    def set_value(self, value):
        self.value = value

    def copy(self):
        return EnvironmentItem(self.name, self.value)


ITEM_TYPE = Union[str, EnvironmentItem]


class EnvironmentVarieble(str):
    def __new__(cls, *args, **kwargs):
        cls = super().__new__(cls)
        return cls

    def __init__(self, name, value_string, split_char=';'):
        self.name = name
        self.split_char = split_char
        self.value = [EnvironmentItem(name, v) for v in _split_env_string(value_string)]

    def append(self, item: ITEM_TYPE):
        if isinstance(item, str):
            item = EnvironmentItem(self.name, item)
        self.value.append(item)

    def remove(self, item: ITEM_TYPE):
        if isinstance(item, str):
            item = EnvironmentItem(self.name, item)
        self.value.remove(item)

    def index(self, item: ITEM_TYPE):
        if isinstance(item, str):
            item = EnvironmentItem(self.name, item)
        return self.value.index(item)

    def pop(self, index: int):
        return self.value.pop(index)

    def copy(self):
        var = EnvironmentVarieble(self.name, '')
        var.value = self.value.copy()
        return var

    def __contains__(self, item: ITEM_TYPE):
        if isinstance(item, str):
            item = EnvironmentItem(self.name, item)
        return item in self.value

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value):
        self.value[index].set_value(value)

    def __iter__(self):
        yield from self.value

    def __str__(self):
        return f'{self.split_char.join([str(v.value) for v in self.value])}'

    __repr__ = __str__


class Environment:
    __slots__ = ('env',)

    def __to_env_dict(self, env):
        result = {}
        for key, value in env._instances():
            if isinstance(value, EnvironmentVarieble):
                result[key] = value
            else:
                result[key] = EnvironmentVarieble(key, value)
        return result

    def __init__(self, env=None):
        if env is None:
            env = os.environ
        elif isinstance(env, Environment):
            env = env.env
        elif isinstance(env, int):
            if env == ENV_ORIGIN:
                env = origin_environ
            elif env == ENV_NEW:
                env = {}
            else:
                raise ValueError(f"{env} is not a valid value")
            env = origin_environ
        else:
            env = dict(env)

        self.env = self.__to_env_dict(env)

    def copy(self):
        return Environment(self.env)

    def keys(self):
        return self.env.keys()

    def values(self):
        return map(str, self.env.values())

    def items(self):
        return map(lambda x: (x[0], str(x[1])), self.env.items())

    def __getitem__(self, key):
        # if inspect.stack()[2].function == 'activate':  # 处理 activate 函数 ( os.update ) 不支持自定义的类型
        #     return str(self.env[key])
        return self.env[key]

    def __setitem__(self, key, value):
        self.env[key] = value

    def __delitem__(self, key):
        del self.env[key]

    def __contains__(self, key):
        return key in self.env

    def __str__(self):
        return f'{get_type_name(self)}({self.env})'

    __repr__ = __str__


class _ManagerEnvironmentItem:
    __slots__ = ('name', 'env')

    def __init__(self, name, value):
        if isinstance(value, str):
            value = EnvironmentVarieble(name, value)
            value = Environment({name: value})

        self.name, self.env = name, value

    def __getitem__(self, item):
        return self.env[item]

    def __setitem__(self, item, value):
        self.env[item] = value

    def __delitem__(self, item):
        del self.env[item]

    @classmethod
    def from_environ(cls, name, *args, **kwargs):
        env = Environment(*args, **kwargs)
        return cls(name, env)

    def copy(self):
        return _ManagerEnvironmentItem(self.name, self.env.copy())

    def activate(self):
        os.environ.update(self.env)

    def avtivate_all(self):
        os.environ.clear()
        self.activate()

    def __str__(self):
        return "Item-" + str(self.env)

    __repr__ = __str__


class EnvironmentManager:
    __slots__ = ('_env',)

    def __init__(self, env_mapping=None):
        if env_mapping is not None:
            self._env: dict[str, _ManagerEnvironmentItem] = {
                k: _ManagerEnvironmentItem(k, v) for k, v in env_mapping._instances()}
        else:
            self._env = {}

    def create_environ(self, name: str, clone_from=None):
        if clone_from is None:
            self._env[name] = _ManagerEnvironmentItem.from_environ(name, os.environ)
        elif isinstance(clone_from, str):
            origin_env = self._env[clone_from]
            self._env[name] = origin_env.copy()
        elif isinstance(clone_from, Environment):
            origin_env = clone_from
            self._env[name] = _ManagerEnvironmentItem(name, origin_env.copy())
        elif isinstance(clone_from, int):
            if clone_from == ENV_ORIGIN:
                self._env[name] = _ManagerEnvironmentItem.from_environ(name, origin_environ)
            elif clone_from == ENV_NEW:
                self._env[name] = _ManagerEnvironmentItem.from_environ(name, {})
            else:
                raise ValueError(f"{clone_from} is not a valid value")
        else:
            raise TypeError(f"{clone_from} is not Envionment item")

    def add_environ(self, name: str, value):
        # 如果Value是Item类型,那么就直接赋值
        # 否则进行类型转换
        self._env[name] = _ManagerEnvironmentItem(name, value) if not isinstance(value,
                                                                                 _ManagerEnvironmentItem) else value

    def env(self, key: str = None, k=None, v=None) -> Union[
        dict[str, _ManagerEnvironmentItem], _ManagerEnvironmentItem]:
        if key is None:  # 返回所有环境变量
            return self._env
        else:
            if k is None:  # 当 k 为 None 时,返回环境变量字典
                return self._env[key]
            elif v is not None:  # 当 k 和 v 都不为 None 时,修改环境变量字典
                self._env[key][k] = v

    def activate(self, key: str):
        self._env[key].activate()

    def activate_all(self, key: str):
        self._env[key].avtivate_all()

    def __getitem__(self, key: str):
        return self._env[key]

    def __setitem__(self, key: str, value):
        if isinstance(value, str):
            value = EnvironmentVarieble(key, value)
            value = Environment({key: value})
            value = _ManagerEnvironmentItem(key, value)

        elif isinstance(value, Environment):
            value = _ManagerEnvironmentItem(key, value)

        elif isinstance(value, EnvironmentVarieble):
            value = Environment({key: value})
            value = _ManagerEnvironmentItem(key, value)

        if not isinstance(value, _ManagerEnvironmentItem):
            raise TypeError(f"{value} is not Envionment item")

        self._env[key] = value

    def __delitem__(self, key: str):
        del self._env[key]



from abc import ABC, abstractmethod
from typing import Any


class ConfigType(ABC):
    value: Any  # 配置项的值
    types: Any  # 配置项值的类型

    parent: Any
    backend: 'BackendABC'

    def __init_subclass__(cls, *, types=None):
        cls.types = types

    def __init__(self, value, parent=None, backend=None):
        self.set(value)
        self.parent = parent
        self.backend = backend

    def transform(self):  # 将配置项的值转换为后端可识别的配置数据
        ...

    @abstractmethod
    def load(self, value):  # 将后端返回的配置数据加载到配置项中
        return value

    @classmethod
    @abstractmethod
    def validate(cls, value) -> bool:  # 检查类型是否符合
        ...

    def set(self, value):
        self.validate(value)
        self.value = value

    def get(self):
        return self.value


from .backend import BackendABC





import time

from .requirements import Task, CallbackMetaclass
from ...model_abc import AbstractDynamicModel


def size_assign(total, max_chunks, chunk_min_size = 1024):
    ans = []
    chunk_size = max(total // max_chunks, chunk_min_size)
    for i in range(max_chunks - 1):
        ans.append((i * chunk_size, (i + 1) * chunk_size - 1))
    ans.append(((max_chunks - 1) * chunk_size, total - 1))
    return ans

def reassign_calc(task: Task):
    speed = task.completed_size / (time.time() - task.start_time)
    total = task.end - task.start + 1
    size = (total - speed * task.connect_time) / 2
    return task.start + int(size)  # 重分配点


class Model(AbstractDynamicModel):
    callback_metaclass: CallbackMetaclass

    def __init__(self, callback_metaclass):
        # super().__init__()
        self._assign = None
        self.total_size = callback_metaclass.get_total_size()
        self.assign_policy = callback_metaclass.get_assign_policy()
        self.tasks = []  # type: list[Task]

    def init(self):
        super().init()
        # Static assign
        self._assign = size_assign(
            self.total_size,
            self.assign_policy.max_chunk_count,
            self.assign_policy.chunk_min_size
        )
        for assign in self._assign:
            task = self.callback_metaclass.create_task(*assign)
            self.tasks.append(task)

    def update(self) -> None:
        for task in self.tasks:
            if task.completed_size == task.end - task.start + 1:
                self.tasks.remove(task)
        if len(self.tasks) == 0:
            self.stop()



# 操作os.pipe()返回的管道

import os
import queue
from collections import deque
from typing import Optional

from .._hycore import threading_methods
from .._hycore.json import Pickle


# module end


class _PIPEWriter:
    """
    Pipe Writer
    """

    def __init__(self, writer: int, buffer_size: int = 1024):
        self._pipe = writer
        self._io = os.fdopen(self._pipe, 'w', buffer_size)

        self._msg_queue = queue.Queue(buffer_size)
        self._event = threading_methods.threading.Event()

        self._write_thread = threading_methods.run_new_daemon_thread(self._write_thread_func)

        self.buffersize = buffer_size

    def set(self, pipe: int):
        self._event.set()
        self._io.close()
        self._write_thread.join()
        self._pipe = pipe
        self._io = os.fdopen(self._pipe, 'w', self.buffersize)
        self._event.clear()
        self._write_thread = threading_methods.run_new_daemon_thread(self._write_thread_func)

    def write(self, data):
        if data is None or data is ...:
            raise ValueError("You should send 'None' or '...' to the reader, they hava their functions.")
        data_text = Pickle.encode(data)
        self._msg_queue.put(data_text)

    def _write_thread_func(self):
        while not self._event.is_set():
            if not self._msg_queue.empty():
                msg = self._msg_queue.get()
                self._io.write(msg)
                self._io.write('\n')
                self._io.flush()


class _PIPEReader:
    """
    Pipe Reader
    """

    def __init__(self, reader: int):
        self._pipe = reader

        self._io = os.fdopen(reader, 'r')

        self._msg_queue = queue.Queue()
        self._event = threading_methods.threading.Event()

        self._read_thread = threading_methods.run_new_daemon_thread(self._read_thread_func)
        self.errors = deque()

    def set(self, reader):
        self._event.set()
        self._read_thread.join()
        self._io.close()
        self._pipe = reader
        self._io = os.fdopen(self._pipe, 'r')
        self._event.clear()
        self._read_thread = threading_methods.run_new_daemon_thread(self._read_thread_func)

    def read(self, timeout: Optional[int] = ...):
        if timeout is ...:
            timeout = None
        try:
            return self._msg_queue.get(timeout=timeout)
        except queue.Empty as e:
            print("EmptyError:", e)
        except TimeoutError as e:
            print("TimeoutError:", e)

    def _read_thread_func(self):
        while not self._event.is_set():
            try:
                one_data = self._io.readline()
                unpacking_data = Pickle.decode(one_data)
                self._msg_queue.put(unpacking_data)
            except Exception as e:
                self.errors.append(e)


def writer(pipe: int, buffer_size: int = 65535):
    return _PIPEWriter(pipe, buffer_size)


def reader(pipe: int):
    return _PIPEReader(pipe)


def get_pair(os_pipe: tuple[int, int], buffer_size: int = 65535):
    return reader(os_pipe[0]), writer(os_pipe[1], buffer_size)



import re
import socket

import ping3


def _ipa():
    """
    返回一个A类IP地址的迭代器
    :return:
    """
    i1 = 10
    for i2 in range(256):
        for i3 in range(256):
            for i4 in range(256):
                yield f"{i1}.{i2}.{i3}.{i4}"


def _ipb():
    """
    返回一个B类IP地址的迭代器
    :return:
    """
    i1 = 172
    for i2 in range(16, 32):
        for i3 in range(256):
            for i4 in range(256):
                yield f"{i1}.{i2}.{i3}.{i4}"


def _ipc():
    """
    返回一个C类IP地址的迭代器
    :return:
    """
    i1 = 192
    i2 = 168
    for i3 in range(256):
        for i4 in range(256):
            yield f"{i1}.{i2}.{i3}.{i4}"


def _ipd():
    """
    返回一个D类IP地址的迭代器
    :return:
    """
    for i1 in range(110, 132):
        for i2 in range(256):
            for i3 in range(256):
                for i4 in range(256):
                    yield f"{i1}.{i2}.{i3}.{i4}"


def _ipe():
    """
    返回一个E类IP地址的迭代器
    :return:
    """
    i1 = 127
    for i2 in range(256):
        for i3 in range(256):
            for i4 in range(256):
                yield f"{i1}.{i2}.{i3}.{i4}"


def _ipf():
    """
    返回一个F类IP地址的迭代器
    :return:
    """
    i1 = 169
    i2 = 254
    for i3 in range(256):
        for i4 in range(256):
            yield f"{i1}.{i2}.{i3}.{i4}"


def _getIP(IPClass: str):
    IPClass = IPClass.upper()[0]
    if IPClass == "A":
        return _ipa()
    elif IPClass == "B":
        return _ipb()
    elif IPClass == "C":
        return _ipc()
    elif IPClass == "D":
        return _ipd()
    elif IPClass == "E":
        return _ipe()
    elif IPClass == "F":
        return _ipf()
    else:
        return None


def getIP(IPClass: str):
    """
    IP范围：\n
    - A: 10.0.0.0-10.255.255.255\n
    - B: 172.16.0.0-172.31.255.255\n
    - C: 192.168.0.0-192.168.255.255\n
    - D: 110.0.0.0-127.255.255.255\n
    - E: 127.0.0.0-127.255.255.255\n
    - F: 169.254.0.0-169.254.255.255\n
    :param IPClass: A, B, C, D, E, F
    :return:
    """
    for i in _getIP(IPClass):
        yield i


def ping(name, timeout=1):
    """
    检查网络连通性
    :param name:
    :param timeout:
    :return:
    """
    return ping3.ping(name, timeout=timeout)


def can_ping(name, timeout=1):
    """
    是否能ping通
    :param name:
    :param timeout:
    :return:
    """
    return bool(ping(name, timeout))


def getIPseq(ip1a, ip1b, ip2a, ip2b, ip3a, ip3b, ip4a, ip4b):
    for ip1 in range(ip1a, ip1b):
        for ip2 in range(ip2a, ip2b):
            for ip3 in range(ip3a, ip3b):
                for ip4 in range(ip4a, ip4b):
                    yield f"{ip1}.{ip2}.{ip3}.{ip4}"


def _getIPv6seq(ip_ranges):
    def parse_hex_range(hex_range: str):
        match = re.match(r'^0x([0-9a-fA-F]+)-0x([0-9a-fA-F]+)$', hex_range)
        if match:
            return int(match.group(1), 16), int(match.group(2), 16)
        else:
            raise ValueError(f"Invalid hexadecimal range: {hex_range}")

    parsed_ranges = [parse_hex_range(hr) for hr in ip_ranges]

    for ip1 in range(parsed_ranges[0][0], parsed_ranges[0][1]):
        for ip2 in range(parsed_ranges[1][0], parsed_ranges[1][1]):
            for ip3 in range(parsed_ranges[2][0], parsed_ranges[2][1]):
                for ip4 in range(parsed_ranges[3][0], parsed_ranges[3][1]):
                    for ip5 in range(parsed_ranges[4][0], parsed_ranges[4][1]):
                        for ip6 in range(parsed_ranges[5][0], parsed_ranges[5][1]):
                            for ip7 in range(parsed_ranges[6][0], parsed_ranges[6][1]):
                                for ip8 in range(parsed_ranges[7][0], parsed_ranges[7][1]):
                                    hex_ip1 = format(ip1, '04x')
                                    hex_ip2 = format(ip2, '04x')
                                    hex_ip3 = format(ip3, '04x')
                                    hex_ip4 = format(ip4, '04x')
                                    hex_ip5 = format(ip5, '04x')
                                    hex_ip6 = format(ip6, '04x')
                                    hex_ip7 = format(ip7, '04x')
                                    hex_ip8 = format(ip8, '04x')

                                    yield \
                                        f"{hex_ip1}:{hex_ip2}:{hex_ip3}:{hex_ip4}:{hex_ip5}:{hex_ip6}:{hex_ip7}:{hex_ip8}"


def getIPv6seq(
        ip1='0x0-0xffff',
        ip2='0x0-0xffff',
        ip3='0x0-0xffff',
        ip4='0x0-0xffff',
        ip5='0x0-0xffff',
        ip6='0x0-0xffff',
        ip7='0x0-0xffff',
        ip8='0x0-0xffff'
):
    return _getIPv6seq([ip1, ip2, ip3, ip4, ip5, ip6, ip7, ip8])


def getIPv6seqi(
        ip1='0-65536',
        ip2='0-65536',
        ip3='0-65536',
        ip4='0-65536',
        ip5='0-65536',
        ip6='0-65536',
        ip7='0-65536',
        ip8='0-65536'
):
    # Convert integer ranges to hexadecimal format for compatibility with _getIPv6seq function
    hex_ip_ranges = [f"0x{ip1[0]:04x}-0x{ip1[1]:04x}",
                     f"0x{ip2[0]:04x}-0x{ip2[1]:04x}",
                     f"0x{ip3[0]:04x}-0x{ip3[1]:04x}",
                     f"0x{ip4[0]:04x}-0x{ip4[1]:04x}",
                     f"0x{ip5[0]:04x}-0x{ip5[1]:04x}",
                     f"0x{ip6[0]:04x}-0x{ip6[1]:04x}",
                     f"0x{ip7[0]:04x}-0x{ip7[1]:04x}",
                     f"0x{ip8[0]:04x}-0x{ip8[1]:04x}"]

    # Validate the input ranges
    for range_str in hex_ip_ranges:
        start, end = range_str.split('-')
        if int(start, 16) < 0 or int(end, 16) > 0xffff:
            raise ValueError(f"Invalid IPv6 sequence integer range: {range_str}")

    return _getIPv6seq(hex_ip_ranges)


def IP4to6(ip4):
    ipv4_bytes = ip4.split('.')

    # 将每个字节转换为十六进制并补足到四位（例如 '0' + '123' -> '0123'）
    ipv4_hex = [hex(int(byte))[2:].zfill(2) for byte in ipv4_bytes]

    # IPv4地址内嵌到IPv6地址的格式为 ::ffff:0a0b:0c0d，其中0a0b:0c0d是IPv4地址的十六进制形式
    ipv6_prefix = "ffff:"
    ipv6_mapped = ipv6_prefix + ":".join(ipv4_hex)

    # 前面添加 "::" 表示高位部分全为0
    full_ipv6_address = "::" + ipv6_mapped

    return full_ipv6_address


def toip(url):
    try:
        ip = socket.gethostbyname(url)
    except socket.gaierror:
        ip = None
    return ip


def toname(ip):
    # 尝试将IP地址解析为域名
    try:
        hostname, _, _ = socket.gethostbyaddr(ip)
        return hostname
    except socket.herror:
        # 如果出现无法反向解析IP地址的情况，返回错误信息
        return None



from . import (
    aes,
    rsa
)


class BothCipher:
    def __init__(self, pubkey, prikey, iv=None):
        self.rsa_pubkey, self.rsa_prikey = pubkey, prikey
        self.aes_iv = iv

    def encrypt(self, plain_text: bytes):
        aes_key = aes.generate_key()
        encode_aes_key = rsa.encrypt(aes_key, self.rsa_pubkey)
        encrypted_data = aes.encrypt(plain_text, aes_key, self.aes_iv)
        return encrypted_data, encode_aes_key

    def decrypt(self, cipher_text: bytes, encoded_aes_key):
        aes_key = rsa.decrypt(encoded_aes_key, self.rsa_prikey)
        return aes.decrypt(cipher_text, aes_key, self.aes_iv)



from typing import Optional

from ..._hycore.type_func import literal_eval as _lt_ev


def literal_eval(string, globals_: Optional[dict] = None, locals_: Optional[dict] = None, builtins: bool = False,
                 no_eval: bool = True):
    return _lt_ev(string, globals_, locals_, builtins, no_eval)



from ...._hyre.re_plus import Literal, Re, BaseRe

NEWLINE = Literal('\n')
IDENT = Re('[a-zA-Z_][a-zA-Z0-9_-]*')
WHITESPACE = Re(r'\s+')
IMPORT = Literal('import')
AS = Literal('as')
FROM = Literal('from')
PASS = Literal('pass')
LP = Re(r'[(\[{]')
RP = Re(r'[)\]}]')
LFILLTOKEN = Literal('{<')
RFILLTOKEN = Literal('>}')
COMMA = Literal(',')
PERIOD = Literal('.')
ASSIGN = Literal('=')
INT = Re('-?[0-9]+')
eINT = Re(r'-?\d+e\d+')
jINT = Re(r'[+-]?[0-9]+j')
hINT = Re(r'0x[0-9a-fA-F]+')
bINT = Re(r'0b[01]+')
oINT = Re(r'0o[0-7]+')
STR = Re(r'"([^"\\]*(\\.[^"\\]*)*)"')
sSTR = Re(r"'([^'\\]*(\\.[^'\\]*)*)'")
multiSTR = Re(r'"""([^"]|"")*"""')
TokenPatterns = [
    ("NEWLINE", NEWLINE),
    # ("DEDENT", ...)  # 退缩记号,由后期添加

    ("KEYWORD", IMPORT),
    ("KEYWORD", AS),
    ("KEYWORD", FROM),
    ("KEYWORD", PASS),

    ("LFILL", LFILLTOKEN),
    ("RFILL", RFILLTOKEN),

    ("OPER", Re(r'((//)|[\+\-\*/^&\|%]|<<|>>)')),

    ("IDENT", IDENT),
    ("ASSIGN", ASSIGN),

    ("INT", INT),
    ("INT", eINT),
    ("INT", jINT),
    ("INT", oINT),
    ("INT", bINT),
    ("INT", hINT),

    ("STR", multiSTR),
    ("STR", STR),
    ("STR", sSTR),

    ("LP", LP),
    ("RP", RP),

    ("COMMA", COMMA),
    ("PERIOD", PERIOD),

    ("WS", WHITESPACE),
    # ("UNKNOWN", ANY),
]  # type: list[tuple[str, BaseRe]]



from .abc import Serializer

# ---

from .S_BinStruct import *
from .S_Json import *
from .S_JsonPickle import *
from .S_Wrapping import *
from .S_Combination import *

# ---


from ._serializer import dumps, loads



from typing import Iterable, Union

from . import hybitmap as _bitmap


class Bitmap(_bitmap.Bitmap):
    def __init__(self, size_or_bits: Union[int, Iterable[bool]]):
        super().__init__(size_or_bits)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.size})'

    def set_bit(self, index: int, value: bool):
        self._C_set_bit(index, value)

    def get_bit(self, index: int):
        self._C_get_bit(index)

    @property
    def size(self):
        return self._C_get_size()

    @property
    def bytes(self):
        return self._C_to_bytes()

    def byte(self, index):
        return self._C_byte_at(index)

    def bbyte(self, byte_index):
        return self._C_byte_at(byte_index * 8)

    __str__ = __repr__



# Safety Evaluation System(SES)
import ast


class SafetyEvaluationSystem:
    def __init__(self, allowed_modules=None, disallowed_nodes=None):
        self.allowed_modules = set(allowed_modules or ['math', 'sys', 'numpy'])
        self.disallowed_nodes = set(disallowed_nodes or [ast.FunctionDef, ast.AsyncFunctionDef])
        self.errors = []
        self.checked_files = set()
        self.safety_score = 100

        self.disallowed_node_score = 5
        self.unsafety_import_score = 10

    def add_allowed_module(self, module_name):
        """Add a module to the allowed list."""
        self.allowed_modules.add(module_name)

    def remove_allowed_module(self, module_name):
        """Remove a module from the allowed list."""
        self.allowed_modules.discard(module_name)

    def add_disallowed_node(self, node_type):
        """Add an AST node type to the disallowed list."""
        self.disallowed_nodes.add(node_type)

    def remove_disallowed_node(self, node_type):
        """Remove an AST node type from the disallowed list."""
        self.disallowed_nodes.discard(node_type)

    def _is_allowed_module(self, module_name):
        """Check if the module name is in the allowed list."""
        return module_name in self.allowed_modules

    def _visit_Import(self, node, filename):
        """Handle import statements."""
        for alias in node.names:
            if not self._is_allowed_module(alias.name):
                self.errors.append(f"Unsafe import detected: {alias.name}")
                self.safety_score -= self.unsafety_import_score  # 每发现一个不安全的导入，减去()分

    def _visit_ImportFrom(self, node, filename):
        """Handle from ... import ... statements."""
        if not self._is_allowed_module(node.module):
            self.errors.append(f"Unsafe import detected: {node.module}")
            self.safety_score -= self.unsafety_import_score  # 同样，不安全的导入减分

    def _check_ast(self, tree, filename):
        """Walk through the abstract syntax tree and check."""
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                self._visit_Import(node, filename)
            elif isinstance(node, ast.ImportFrom):
                self._visit_ImportFrom(node, filename)
            elif type(node) in self.disallowed_nodes:
                self.errors.append(f"Disallowed node found: {type(node).__name__}")
                self.safety_score -= self.disallowed_node_score  # 发现不允许的节点类型，减去5分
            else:
                # Only allow imports and class definitions by default
                pass

    def _check_file(self, filename):
        """Recursively check the safety of a file and its dependencies."""
        if filename in self.checked_files:
            return
        self.checked_files.add(filename)
        with open(filename, 'r') as file:
            source = file.read()
            tree = ast.parse(source)
            self._check_ast(tree, filename)

    def check_file(self, filename):
        """Check the safety of the given file and its dependencies."""
        self._check_file(filename)
        return self.safety_score, self.errors



import time
from typing import Optional

from .process import CProcess as _CProcess


class Process(_CProcess):
    def pause(self):
        self.suspend()

    def recover(self):
        self.resume()

    def exitcode(self) -> Optional[int]:
        return self.wait()


class Timer:
    def __init__(self):
        self._time = 0
        self._res = 0

    def start(self):
        self._time = time.time()

    def stop(self):
        self._res = time.time() - self._time
        return self._res

    def reset(self):
        self._time = 0
        self._res = 0

    @property
    def res(self):
        return self._res



import rsa


# module end

def generate_keys(size: int = 2048):
    """
    return pubkey, privkey
    """
    return rsa.newkeys(size)


def encrypt(plain_text: bytes, key: rsa.PublicKey):
    """
    return cipher_text
    """
    return rsa.encrypt(plain_text, key)


def decrypt(cipher_text: bytes, key: rsa.PrivateKey):
    """
    return plain_text
    """
    return rsa.decrypt(cipher_text, key)



import logging
import os
from typing import Iterable

import pyaes

from .func import split, pad, unpad

zencrypt_logger = logging.getLogger("SzQlib.zencrypt")


# module end


def aes_encrypt(plaintext: bytes, key, iv=None):
    if iv is None:
        iv = os.urandom(16)

    plaintext_bytes = plaintext
    padded_plaintext = plaintext_bytes
    aes = pyaes.AESModeOfOperationCBC(key, iv)
    ciphertext = aes.encrypt(padded_plaintext)
    return ciphertext, iv


def aes_decrypt(ciphertext, key, iv):
    aes = pyaes.AESModeOfOperationCBC(key, iv)
    decrypted_data = aes.decrypt(ciphertext)
    return decrypted_data


def aes_encrypt_ls(plain_text_list: Iterable, key, vi=None, NoIV=False):
    for i in plain_text_list:
        if NoIV:
            yield aes_encrypt(i, key, vi)[0]
        else:
            yield aes_encrypt(i, key, vi)


def aes_detrypt_ls(cipher_text_list, key, vi):
    for i in cipher_text_list:
        yield aes_decrypt(i, key, vi)


def generate_key(size=16):
    return os.urandom(size)


def generate_iv(size=16):
    return os.urandom(size)


def generate(size=16):
    return generate_key(size), generate_iv(size)


def split_pad(text: bytes, size=16):
    return [pad(i, size) for i in split(text, size)]


def join(args):
    return b''.join(args)


def encode(text: bytes, key, iv):  # 提供一个最简单的加密api
    """
    加密函数，提供一个最简单的api加快使用效率
    :param text:
    :param key:
    :param iv:
    :return: bytes
    """
    byte = text
    byte = split_pad(byte)
    crypt_text = aes_encrypt_ls(byte, key, iv, NoIV=True)

    return join(crypt_text)


def decode(crypt_text: bytes, key, iv):  # 提供一个最简单的解密api
    """
    解密，提供一个最简单的api加快使用效率
    :param crypt_text:
    :param key:
    :param iv:
    :return: str
    """
    crypt_list = split(crypt_text)
    plain_text = list(aes_detrypt_ls(crypt_list, key, iv))
    for i in range(len(plain_text)):
        plain_text[i] = unpad(plain_text[i])
    return join(plain_text)


def encrypt(plain_text: bytes, key, iv=None):  # 别名
    return encode(plain_text, key, iv)


def decrypt(crypt_text: bytes, key, iv=None):  # 别名
    return decode(crypt_text, key, iv)


zencrypt_logger.debug("Module zencrypt loading ...")



from abc import ABC, abstractmethod


class REConcater(ABC):
    @classmethod
    @abstractmethod
    def concat(cls) -> str: ...
    @abstractmethod
    def pattern(self) -> str: ...

    @abstractmethod
    def __add__(self, other): ...

    @abstractmethod
    def __mul__(self, other): ...

    @abstractmethod
    def __or__(self, other): ...


class BaseREConcateratable(ABC):
    name = None
    ignore = False
    origin_pattern = None

    @abstractmethod
    def pattern(self) -> str: ...
    @abstractmethod
    def origin(self) -> str: ...

    def findall(self, string): ...

    def finditer(self, string): ...

    def match(self, string): ...



from .model_abc import *
from ...hycore import Clock


class Runner:
    """
    将模型和执行器绑定,按照指定频率刷新模型
    """
    def __init__(self, model: AbstractDynamicModel = None, executor: AbstractExecutor = None):
        self.model = model
        self.executor = executor
        self.clock = Clock()
        self._running = True

    def set_model(self, model: AbstractDynamicModel):
        self.model = model

    def set_executor(self, executor: AbstractExecutor):
        self.executor = executor

    def stop(self):
        self._running = False
        self.executor.on_stop()

    def start(self, interval: float = 0.1):
        self.executor.on_run()
        while self._running:
            self.model.update()
            self.clock.strike(interval)
            if self.model.running is False:
                self.stop()



from typing import Union

from . import _bitmap


class Bitmap(_bitmap.Bitmap):
    def __init__(self, size: int = 1):
        super().__init__(size)

    @property
    def size(self) -> int:
        return self._C_get_size()

    def set(self, index: int, value: Union[bool, int]):
        return self._C_set_bit(index, value)

    def get(self, index: int) -> bool:
        return self._C_get_bit(index)

    def __iter__(self):
        for i in range(self.size * 8):
            yield self.get(i)

    def to_bytes(self):
        return self._C_to_bytes()

    def byte_at(self, index):
        return self._C_byte_at(index)

    def __getitem__(self, index: int) -> bool:
        return self.get(index)

    def __setitem__(self, index: int, value: Union[bool, int]):
        return self.set(index, value)

    def __len__(self):
        return self.size * 8

    def __str__(self):
        return self.to_bytes().hex()

    __repr__ = __str__



from ._block_parser import \
    BlockParser

from ._syntax_parser import \
    SyntaxParser

from ._types import \
    Block

from ..Lexer import \
    Lexer


class Parser:

    def __init__(self):
        self._lexer = Lexer()
        self._block_parser = BlockParser()
        self._syntax_parser = SyntaxParser()
        self.tokens = None
        self.block = None
        self.ast = None

    def parse(self, source_code):
        self.tokens = self._lexer.lex(source_code)
        self.block = self._block_parser.parse(self.tokens)
        self.ast = self._syntax_parser.parse(self.block)



import inspect
import os

import psutil

from .process_addons import Process

if os.name == 'nt':

    def shutdown(mode: str, _time: int = 1):
        if _time < 1:
            _time = 1
        return os.system(f'shutdown -{mode} -t {_time}')


class Runtime:
    _instance = None

    @classmethod
    def __getruntime(cls, *args, **kwargs):
        return cls.__new__(cls, *args, **kwargs)

    def __new__(cls, *args, **kwargs):
        stact = inspect.stack()[1].function
        if stact != "__getruntime":
            raise RuntimeError("Runtime class can only be instantiated by the getruntime method")
        if cls._instance is None:
            cls._instance = super(Runtime, cls).__new__(cls)

        return cls._instance

    def __init__(self):
        self._c = Process(os.getpid())

    @property
    def pid(self):
        return self._c.pid

    def cpu_count(self):
        return os.cpu_count()

    def num_threads(self):
        return self._c.num_threads()

    def cpu_percent(self):
        return self._c.cpu_percent()

    def memory_percent(self):
        return self._c.memory_percent()

    def memory_info(self):
        return self._c.memory_info()

    def memory_full_info(self):
        return self._c.memory_full_info()

    def kill(self):
        self._c.kill()

    def exit(self, code):
        exit(code)

    def environ(self):
        return self._c.environ()

    def cmdline(self):
        return self._c.cmdline()

    def exec(self, command):
        return psutil.Popen(command)

    def execute(self, command):
        return Process(psutil.Popen(command).pid)

    @classmethod
    def getRuntime(cls):
        return cls.__getruntime()



import fractions
from collections import deque
from collections.abc import Iterable


def d1(value, lenght):
    return [value for _ in range(lenght)]


def d2(value, lenght, width):
    return [d1(value, width) for _ in range(lenght)]


def d3(value, lenght, width, height):
    return [d2(value, lenght, width) for _ in range(height)]


def d1_init(ls, value):
    lenght = len(ls)
    return d1(value, lenght)


def d2_init(ls, value):
    lenght = len(ls)
    width = len(ls[0])
    return d2(value, lenght, width)


def d3_init(ls, value):
    lenght = len(ls)
    width = len(ls[0])
    height = len(ls[0][0])
    return d3(value, lenght, width, height)


def sub(list1, list2):
    # is lst1 in lst2
    return list1 in list2


def parent(list1, list2):
    # is lst1 of lst2
    return list2 in list1


def match(list1, list2):
    return list1 == list2


def indexs_of(lst, indexs=None):
    if indexs is None:
        return []
    return [lst[i] for i in indexs]


def hasindex(iterable, index):
    if isinstance(index, slice):
        return True  # Slice 不会引发错误，不需要检查
    if index < 0:
        index += len(iterable)
    return len(iterable) > index


def split(ls, split_nums: Iterable[int]):
    """
    将列表按照**整数**比例分割，返回分割后的列表
    如:
        a = [1, 2, 3, 4, 5, 6]
        b, c, d = split(a, 1, 2, 3)  # b: [1], c: [2, 3], d: [4, 5, 6]
    """
    sm = sum(split_nums)
    lengths = [fractions.Fraction(i) * sm for i in split_nums]
    cursor = 0
    for l in lengths:
        yield [i for i in ls[cursor:cursor + l]]
        cursor += l


class _ListConcater:
    """
    逻辑连接两个列表
    """

    def __init__(self, *ls):
        self.ls_list = ls
        self.lg_list = []

        self.flush()

    def flush(self):
        self.lg_list = [len(i) for i in self.ls_list]

    def _find_list(self, idx, num):
        if num < 0:
            num += sum(self.lg_list)
        if idx >= len(self.lg_list):
            raise IndexError('index out of the range')
        if num >= self.lg_list[idx]:
            return self._find_list(idx + 1, num - self.lg_list[idx])
        return idx, num

    def _get(self, idx):
        last_idx, list_idx = self._find_list(0, idx)
        return self.ls_list[last_idx][list_idx]

    def _set(self, idx, value):
        last_idx, list_idx = self._find_list(0, idx)
        self.ls_list[last_idx][list_idx] = value

    def _get_range_length(self, start, stop, step):
        return (stop - start) // step

    def append(self, v):
        self.ls_list[-1].append(v)
        self.flush()

    def extend(self, v):
        self.ls_list[-1].extend(v)
        self.flush()

    def list(self):
        return [
            item for ls in self.ls_list for item in ls
        ]

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._get(key)

        if isinstance(key, slice):
            start = key.start or 0
            stop = key.stop or len(self)
            step = key.step or 1
            return [self._get(i) for i in range(start, stop, step)]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self._set(key, value)

        if isinstance(key, slice):
            start = key.start or 0
            stop = key.stop or len(self)
            step = key.step or 1
            if not len(value) == self._get_range_length(start, stop, step):
                raise ValueError('length of value is not equal to the range')
            for si, oi in zip(range(start, stop, step), range(len(value))):
                self._set(si, value[oi])

    def __len__(self):
        return sum(self.lg_list)


def concat(*ls):
    return _ListConcater(*ls)


class _ListFillConcater:
    """
    以填充覆盖方式连接两个列表，如:
        a = [1, 2, 3, 4, 5, 6]
        b = [5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        fill_concat_ls = fill_concat(a, b)  # [1, 2, 3, 4, 5, 6, 11, 12, 13, 14]
        其中a列表被逻辑覆写进了b列表的开头
        如果fill_ls的长度大于main_ls，那么无论怎样访问，都将优先返回fill_ls的值，但是**合并长度不改变**
        列表的元数据以main_ls为基准
        len(fill_concat_ls)  # 10
    """

    def __init__(self, fill_ls, main_ls):
        self.ls_fill, self.main_ls = fill_ls, main_ls

    def _get(self, item):
        if item >= len(self):
            raise IndexError('index out of the range')
        if hasindex(self.ls_fill, item):
            return self.ls_fill[item]
        if not hasindex(self.main_ls, item):
            raise IndexError('index out of the range')
        return self.main_ls[item]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get(item)

        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or len(self)
            step = item.step or 1
            return [self._get(i) for i in range(start, stop, step)]

    def list(self):
        return [
            self[i] for i in range(len(self))
        ]

    def __setitem__(self, key, value):
        raise NotImplementedError('Fill concat cannot be modified')

    def __len__(self):
        return len(self.main_ls)


def fill_concat(fill_ls, main_ls):
    return _ListFillConcater(fill_ls, main_ls)


class _ListReplaceConcater:
    class ReplaceIndex:
        def __init__(self, value, length):
            self.value = value
            self.length = length

    def __init__(self, main_ls):
        self.main_ls = main_ls
        self.replace_ls = []
        self.fore_sum = deque([0])  # 前缀和

    def list(self):
        return [
            self._get(i) for i in range(len(self)-1)
        ]

    def _replace_length(self):
        return self.fore_sum[-1]

    def replace_one(self, value, length):
        """
        将一个值作为逻辑代替项代替main_ls中的length个项
        """
        self.replace_ls.append(self.ReplaceIndex(value, length))
        self.fore_sum.append(
            self.fore_sum[-1] + length - 1
        )

    def _get(self, item):
        if item >= len(self):
            raise IndexError('index out of the range')
        if item < len(self.replace_ls):
            return self.replace_ls[item].value
        else:
            return self.main_ls[item + self._replace_length()]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get(item)
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or len(self)
            step = item.step or 1
            return [self._get(i) for i in range(start, stop, step)]

    def __len__(self):
        s = sum(i.length - 1 for i in self.replace_ls)
        return len(self.main_ls) - s  # 减去逻辑替换项的长度


def replace_concat(main_ls):
    return _ListReplaceConcater(main_ls)



import asyncio
from typing import AsyncGenerator

from ..._hycore.socket import Asyncsocket
from ..._hycore.type_func import pack_variable_length_int, unpack_variable_length_int


async def sock_to_iterable(s: Asyncsocket):
    while True:
        yield await s.recv(1)


def sync_iter(agr: AsyncGenerator):
    loop = asyncio.get_running_loop()
    try:
        while True:
            yield loop.run_until_complete(agr.__anext__())
    except (StopAsyncIteration, asyncio.CancelledError, GeneratorExit):
        loop.run_until_complete(agr.aclose())


def get_part_from_sock(s: Asyncsocket):
    sock_generator = sync_iter(sock_to_iterable(s))
    data, _ = unpack_variable_length_int(
        sock_generator
    )
    sock_generator.close()
    return data


async def build_part_into_sock(data: bytes, s: Asyncsocket):
    length = len(data)
    await s.sendall(
        pack_variable_length_int(length) + data
    )



class Range:
    def __init__(self, start, end, step=1):
        self.start = start
        self.end = end
        self.step = step

    def __iter__(self):
        yield from range(self.start, self.end, self.step)



from collections import deque
from typing import Iterable

from . import Tree, HuffmanTree, Graph, BplusTree


class Stack:
    @property
    def ls(self):
        return tuple(self._ls)

    @ls.setter
    def ls(self, ls):
        self._ls = deque(ls)

    def __init__(self, lst: Iterable = None):
        self._ls = deque() if lst is None else deque(lst)

    def push(self, data):
        """
        Push data to stack.
        """
        self._ls.append(data)

    def pop(self):
        """
        Pop data from stack(delete last one).
        if stack is empty, return None.
        """
        return self._ls.pop() if not self.is_empty() else None

    def size(self):
        """
        Get stack size.
        """
        return len(self._ls)

    def is_empty(self):
        """
        Check stack is empty.
        """
        return self.size() == 0

    def top(self):
        """
        Get stack top data.
        """
        return self._ls[-1] if not self.is_empty() else None

    def copy(self):
        """
        Copy stack.
        """
        return self.__class__(self._ls.copy())

    @property
    def at_top(self):
        """
        Same as `.top()`.
        """
        return self.top()

    @at_top.setter
    def at_top(self, data):
        if not self.is_empty():
            self._ls[-1] = data

    def __str__(self):
        return str(self._ls)

    def __iter__(self):
        return iter(self._ls)

    def __getitem__(self, item):
        return self._ls[item]

    def __len__(self):
        return len(self._ls)

    def __setitem__(self, key, value):
        self._ls[key] = value

    def __repr__(self):
        return "Stack({self.lst})".format(self=self)


class Heap:
    """
    默认大顶堆
    """
    def __init__(self, lst=None, reversed=False):
        self.heap = lst or []
        self.iter_reversed = reversed

    def _parent(self, i):
        return (i - 1) // 2

    def _left_child(self, i):
        return 2 * i + 1

    def _right_child(self, i):
        return 2 * i + 2

    def _swap(self, i, j):
        self.heap[i], self.heap[j] = self.heap[j], self.heap[i]

    def _heapify_down(self, index):
        left = self._left_child(index)
        right = self._right_child(index)
        smallest = index

        if left < len(self.heap) and self.heap[left] < self.heap[smallest]:
            smallest = left

        if right < len(self.heap) and self.heap[right] < self.heap[smallest]:
            smallest = right

        if smallest != index:
            self._swap(index, smallest)
            self._heapify_down(smallest)

    def _heapify_up(self, index):
        parent = self._parent(index)
        if index > 0 and self.heap[parent] > self.heap[index]:
            self._swap(index, parent)
            self._heapify_up(parent)

    def insert(self, value):
        self.heap.append(value)
        self._heapify_up(len(self.heap) - 1)

    def remove(self, value):
        index = self.heap.index(value)
        self._swap(index, len(self.heap) - 1)
        self.heap.pop()
        self._heapify_down(index)

    def append(self, value):
        self.insert(value)

    def extract_min(self):
        if not self.heap:
            raise IndexError("Heap is empty.")
        min_val = self.heap[0]
        last_val = self.heap.pop()
        if self.heap:
            self.heap[0] = last_val
            self._heapify_down(0)
        return min_val

    def peek(self):
        if not self.heap:
            raise IndexError("Heap is empty.")
        return self.heap[0]

    def copy(self):
        return self.__class__(self.heap.copy())

    def iter(self):
        copy = self.copy()
        while copy:
            yield copy.heap[0]
            last = copy.heap.pop()
            if copy:
                copy.heap[0] = last
                copy._heapify_down(0)

    iter_reversed = False

    def __iter__(self):
        return self.iter() if not self.iter_reversed else reversed(tuple(self.iter()))

    def __len__(self):
        return len(self.heap)



def sub(dic1, dic2):
    # dic1 in dic2
    if tuple(dic1) <= tuple(dic2):
        for i in dic1:
            if i not in dic2 or dic1[i] != dic2[i]:
                return False
        return True
    else:
        return False


def parent(dic1, dic2):
    # dic2 in dic1
    return sub(dic2, dic1)


def key_sub(dic1, dic2):
    return tuple(dic1) <= tuple(dic2)


def key_parent(dic1, dic2):
    return tuple(dic2) <= tuple(dic1)


def update(dic1, dic2):
    # update 将dic2 的值更新到 dic1
    for i in dic2:
        dic1[i] = dic2[i]
    return dic1


def eupdate(dic1, dic2):
    for i in dic2:
        if i not in dic1:
            dic1[i] = dic2[i]
    return dic1


def sort(dic1, dic2):
    """
    将dic1和dic2的键修改至相同（key），返回dic1
    当dic1.keys() == dic2.keys()，字典不会发生任何更改
    """
    # 将两部字典的键修改至相同
    for i in dic1:
        if i not in dic2:
            dic2[i] = dic1[i]
    for i in dic2:
        if i not in dic1:
            dic1[i] = dic2[i]
    return dic1


def get_pairs_by_value(dic, value, reverse=False):
    """
    通过值获取所以符合的键值对，结果以字符排序（默认升序），可以通过reverse参数控制

    :param dic: a dict type object
    :param value: a value of dict
    :param reverse: 控制升序降序，对于sort/sorted中的reverse参数
    """
    pairs = [(k, v) for k, v in dic._instances() if v == value]
    return sorted(pairs, key=lambda x: x[0], reverse=reverse)


class AttrDict:
    def __init__(self, **kwargs):
        object.__setattr__(self, "_dict", kwargs)

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __getitem__(self, item):
        return self._dict[item]

    def __setitem__(self, key, value):
        self._dict[key] = value

    @property
    def get_dict(self):
        return self._dict


from types import MappingProxyType


class ConstDict:
    def __init__(self, initial_dict=None):
        if initial_dict is None:
            initial_dict = {}
        self._data = MappingProxyType(initial_dict)

    def __getitem__(self, key):
        return self._data[key]

    def get(self, key, *, default=None):
        return self._data.get(key, default)

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def __repr__(self):
        return f"ConstDict({self._data})"



import struct
from decimal import Decimal

from .type_func import pack_variable_length_int, unpack_variable_length_int


def neopack(data):
    """
    对基本Struct模块的功能进行一定的改进的方法
    **Format Mro**

    - x           pad byte
    - c           char
    - b           int8
    - B           uint8
    - h           int16
    - H           uint16
    - i           int32
    - I           uint32
    - l           int64
    - L           uint64
    - f           float
    - d           double
    - s           char[]
    - q           long long
    - Q           unsigned long long
    - e           half float
    - f           float
    - d           double
    - n           long long
    - N           unsigned long long
    - p           bytes
    - P           int(Point)
    - ?           bool
    """
    data_type = type(data)
    if data_type == int:
        return pack_variable_length_int(data)
    elif data_type in [float, Decimal]:
        # 0.00001
        count = 0
        while data - int(data) != 0:
            data *= 10
            count += 1
        part1 = pack_variable_length_int(count)
        part2 = pack_variable_length_int(int(data))
        return part1 + part2
    elif data_type == str:
        return data.encode()
    elif data_type == bytes:
        return data
    elif data_type == bool:
        return struct.pack("<?", data)
    else:
        raise TypeError("unsupported data type: {}".format(data_type))


def neounpack(data_type, data):
    if data_type == int:
        return unpack_variable_length_int(data)[0]
    elif data_type in [float, Decimal]:
        offset, length = unpack_variable_length_int(data)
        bytes_, _ = unpack_variable_length_int(data[length:])
        print(bytes_)
        return data_type(bytes_) / (10 ** offset)
    elif data_type == str:
        return data.decode()
    elif data_type == bytes:
        return data
    elif data_type == bool:
        return struct.unpack("<?", data)[0]
    else:
        raise TypeError("Unsupported data type: {}".format(data_type))



# ast types
from src.HydrogenLib.TypeFunc.LiteralEval import literal_eval


class Expr:
    def __init__(self, expr):
        self.expr = expr

    def execute(self, globals_: dict | None = None, locals_: dict | None = None):
        return literal_eval(self.expr, globals_, locals_, builtins=True)


class Table:
    def __init__(self, attrs: dict):
        self.attrs = attrs

    def add_attr(self, name, value):
        self.attrs[name] = value

    def delete_attr(self, name):
        del self.attrs[name]

    def get_attr(self, name):
        return self.attrs[name]

    def exists(self, name):
        return name in self.attrs


class Var:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def set(self, value):
        self.value = value

    def get(self):
        return self.value



from typing import Any, Union
from collections import UserDict


class IDict_Item:
    def __init__(self, ins, value):
        self.ins = ins
        self.value = value


class InstanceDict(UserDict):
    def __init__(self, dct=None):
        super().__init__()
        if dct:
            for k, v in dct._instances():
                self._set(k, v)

    def _get(self, key):
        return super().__getitem__(key)

    def _set(self, key, value):
        super().__setitem__(id(key), IDict_Item(key, value))

    def get(self, k, id_key=False, default=None, item=False) -> Union[IDict_Item, Any]:
        if not id_key:  # 如果 k 不作为 id 传入
            k = id(k)  # 转换为 id

        if k not in self:  # 如果 k 不位于字典中
            return default  # 返回默认值

        if item:  # 返回 IDict_Item
            return self._get(k)
        else:  # 返回 value
            return self._get(k).value

    def set(self, k, v):
        self._set(k, v)

    def delete(self, key):
        del self[id(key)]

    def pop(self, key, id_key=False):
        if not id_key:
            key = id(key)

        return super().pop(key)

    def __getitem__(self, key):
        return self._get(id(key))

    def __setitem__(self, key, value):
        self._set(key, value)

    def __delitem__(self, key):
        super().__delitem__(id(key))
        
    def __contains__(self, item):
        return super().__contains__(id(item))



import asyncio
from functools import partial

from .threading_methods import run_new_thread


def run_in_new_loop(coro):
    loop = asyncio.new_event_loop()
    return loop, loop.run_until_complete(coro)


def run_in_existing_loop(coro, loop: asyncio.AbstractEventLoop):
    return loop.run_until_complete(coro)


def new_event_loop():
    return asyncio.new_event_loop()


async def to_coro(func, *args, **kwargs):
    return func(*args, **kwargs)


class _Wrap:
    def __init__(self, func, *args, **kwargs):
        self._partial = partial(func, *args, **kwargs)
        self._future = asyncio.Future()

    def __call__(self, *args, **kwargs):
        self._future.set_result(self._partial(*args, **kwargs))

    @property
    def done(self):
        return self._future.done()

    @property
    def result(self):
        return self._future.result()

    def __await__(self):
        return self._future.__await__()


def wrap(func, *args, **kwargs):
    return _Wrap(func, *args, **kwargs)


class ProtectedTask:
    def __init__(self, loop, origin_task_instance):
        self._loop: asyncio.AbstractEventLoop = loop
        self._task: asyncio.Task = origin_task_instance

    def done(self):
        return self._task.done()

    def cancel(self):
        self._loop.call_soon_threadsafe(self._task.cancel, ())

    def __await__(self):
        return self._task.__await__()


class ThreadEventLoop:
    """
    独立运行的事件循环,保证线程安全
    """

    def __init__(self):
        self._thread = None
        self._loop: asyncio.AbstractEventLoop = None

    def __run_threadsafe(self, func, *args, **kwargs):
        wraped_func = wrap(func, *args, **kwargs)
        self._loop.call_soon_threadsafe(wraped_func)
        return wraped_func.result

    def __thread_main(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def start(self):
        self._loop = new_event_loop()
        self._thread = run_new_thread(self.__thread_main)

    def create_task(self, coro):
        task = self.__run_threadsafe(self._loop.create_task, coro)
        return ProtectedTask(self._loop, task)

    def run_until_complete(self, coro):
        return asyncio.run_coroutine_threadsafe(coro, self._loop).result()

    def run_coroutine(self, coro):
        return asyncio.run_coroutine_threadsafe(coro, self._loop)

    def all_task(self):
        return asyncio.all_tasks(self._loop)

    def stop(self):
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join()






from typing import Iterable, Union, Optional

from bitarray import bitarray


class Bitmap:
    def __init__(self, bits: Optional[Union[int, Iterable[int]]] = None):
        """
        初始化 Bitmap 对象。

        :param bits: 可以是整数、可迭代的位序列或 None。
                     如果是整数，则将其作为位图的初始值。
                     如果是可迭代的位序列，则将其转换为位图。
                     如果是 None，则初始化为空位图。
        """
        self.bits = bitarray()
        if bits is None:
            self.bits = bitarray()
        elif isinstance(bits, int):
            self.bits = bitarray(format(bits, 'b'))
        elif isinstance(bits, Iterable):
            self.bits = bitarray(map(int, bits))
        else:
            raise TypeError("bits must be an int, an iterable of bits, or None")

    def get_bit(self, index: int) -> bool:
        """
        获取指定索引的位值。

        :param index: 位的索引。
        :return: 指定索引的位值（True 或 False）。
        """
        if index >= len(self.bits):
            return False
        return bool(self.bits[index])

    def set_bit(self, index: int, on: bool = True):
        """
        设置指定索引的位值。

        :param index: 位的索引。
        :param on: 如果为 True，则将位设置为 1；否则设置为 0。
        """
        if index >= len(self.bits):
            self.bits.extend([False] * (index - len(self.bits) + 1))
        self.bits[index] = on

    def pack(self) -> bytes:
        """
        将位图打包为字节序列。

        :return: 字节序列。
        """
        return self.bits.tobytes()

    @classmethod
    def unpack(cls, data: bytes) -> 'Bitmap':
        """
        从字节序列解包为 Bitmap 对象。

        :param data: 字节序列。
        :return: Bitmap 对象。
        """
        bits = bitarray()
        bits.frombytes(data)
        return cls(bits)

    def extend(self, other: 'Bitmap'):
        """
        将另一个 Bitmap 对象扩展到当前 Bitmap 对象中。

        :param other: 另一个 Bitmap 对象。
        """
        self.bits.extend(other.bits)

    def __setitem__(self, key: int, value: bool):
        """
        设置指定索引的位值。

        :param key: 位的索引。
        :param value: 位的值（True 或 False）。
        """
        self.set_bit(key, value)

    def __getitem__(self, item: int) -> bool:
        """
        获取指定索引的位值。

        :param item: 位的索引。
        :return: 指定索引的位值（True 或 False）。
        """
        return self.get_bit(item)

    def __iter__(self):
        """
        返回一个迭代器，用于遍历位图中的每一位。

        :return: 迭代器。
        """
        return iter(self.bits)

    def __len__(self) -> int:
        """
        返回位图的长度（即位的数量）。

        :return: 位图的长度。
        """
        return len(self.bits)

    def __str__(self) -> str:
        """
        返回位图的字符串表示。

        :return: 字符串表示。
        """
        return f"{self.__class__.__module__}.{self.__class__.__name__}({self.bits.to01()})"

    __repr__ = __str__



import time
from typing import Union


class Vtime:
    def __init__(self, sec):
        self._sec = sec
        self._hor = 0
        self._min = 0
        self._day = 0

        self.process()

    def process(self):
        flag = 0
        while not flag:
            flag = 1
            if self._sec >= 60:
                self._min += self._sec // 60
                self._sec -= self._sec // 60 * 60
                flag = 0
            if self._min >= 60:
                self._hor += self._min // 60
                self._min -= self._min // 60 * 60
                flag = 0
            if self._hor >= 24:
                self._day += self._hor // 24
                self._hor -= self._hor // 24 * 24
                flag = 0

        self._min = int(self.min)
        self._hor = int(self.hor)
        self._day = int(self.day)

    def toInt(self):
        self._day = int(self.day)
        self._hor = int(self.hor)
        self._sec = int(self.sec)
        self._min = int(self.min)

    @property
    def sec(self):
        return self._sec

    @property
    def min(self):
        return self._min

    @property
    def hor(self):
        return self._hor

    @property
    def day(self):
        return self._day

    @sec.setter
    def sec(self, v: Union[int, float]):
        self._sec = v

    @min.setter
    def min(self, v: Union[int, float]):
        self._min = v

    @hor.setter
    def hor(self, v: Union[int, float]):
        self._hor = v

    @day.setter
    def day(self, v: Union[int, float]):
        self._day = v

    @property
    def time(self):
        return self.day, self.hor, self.min, self.sec

    @time.setter
    def time(self, v: Union[tuple[int | float, ...], list[int | float, ...]]):
        lenght = len(v)
        if lenght >= 1:
            self.day = v[0]
        if lenght >= 2:
            self.hor = v[1]
        if lenght >= 3:
            self.min = v[2]
        if lenght >= 4:
            self.sec = v[3]

    def get_sec(self):
        return self.day * 24 * 60 * 60 + self.hor * 60 * 60 + self.min * 60 + self.sec

    def __str__(self):
        return f"{self.__class__.__name__}({self.get_sec()})"

    __repr__ = __str__


class Timer:
    def __init__(self):
        self.start_time = None
        self.end_time = None
        self.res = None

    def start(self):
        self.start_time = time.time()

    def end(self):
        if self.start_time is None:
            return None
        self.end_time = time.time()
        res = Vtime(self.end_time - self.start_time)
        self.res = res
        return res



from types import NoneType
from typing import Literal

from .BinaryTree import *
from .Bitmap import *
from .Dict import *
from .Func import *
from .IndexOffset import *
from .List import *
from .LiteralEval import *
from .Number import *
from .Overload import overload, get_func_overloads
from .Template import *
from .Type import *

builtin_types = Union[int, float, str, bool, NoneType, list, tuple, dict, set, bytes, bytearray]


def int_to_bytes(num: int, lenght, byteorder: Literal["little", "big"] = 'little'):
    try:
        byte = num.to_bytes(lenght, byteorder)
    except OverflowError as e:
        raise e
    return byte


def int_to_bytes_nonelength(num: int):
    length = len(hex(num))
    length = max(length // 2 - 1, 1)  # 十六进制下,每两个字符占一个字节
    return num.to_bytes(length, 'little')


def bytes_to_int(data: bytes, byteorder: Literal["little", "big"] = 'little'):
    if len(data) == 0:
        return
    return int.from_bytes(data, byteorder)


def get_vaild_data(data: bytes) -> bytes:
    """
    100100 -> vaild = 1001
    111100 -> vaild = 1111
    :param data:
    :return:
    """
    # 找到第一个有效数据（逆序），他的后面就是无效数据
    after_data = data[::-1]
    for index, value in enumerate(after_data):
        if value != 0:
            last_invaild = len(data) - index
            return data[:last_invaild]
    return b''


def is_errortype(exception) -> bool:
    return isinstance(exception, Exception)


def get_attr_by_path(path, Globals=None, Locals=None):
    """
    :param path: 引用路径
    :param Globals: globals() 字典
    :param Locals: locals() 字典

    """
    try:
        return LiteralEval.literal_eval(path, Globals, Locals)
    except (NameError, ValueError, SyntaxError, TypeError, Exception):
        return None


def get_type_name(origin_data):
    return origin_data.__class__.__name__



import os
import sys
import types
import winreg


def add_to_startup(name, file_path=""):
    # By IvanHanloth
    if file_path == "":
        file_path = os.path.realpath(sys.argv[0])
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"Software\Microsoft\Windows\CurrentVersion\Run",
                         winreg.KEY_SET_VALUE,
                         winreg.KEY_ALL_ACCESS | winreg.KEY_WRITE | winreg.KEY_CREATE_SUB_KEY)  # By IvanHanloth
    winreg.SetValueEx(key, name, 0, winreg.REG_SZ, file_path)
    winreg.CloseKey(key)


def remove_from_startup(name):
    key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, r"Software\Microsoft\Windows\CurrentVersion\Run",
                         winreg.KEY_SET_VALUE,
                         winreg.KEY_ALL_ACCESS | winreg.KEY_WRITE | winreg.KEY_CREATE_SUB_KEY)  # By IvanHanloth
    try:
        winreg.DeleteValue(key, name)

    except FileNotFoundError:
        print(f"{name} not found in startup.")
    else:
        print(f"{name} removed from startup.")

    winreg.CloseKey(key)


def value_HKEY(name):
    reg_path_l = name
    reg = winreg.HKEY_LOCAL_MACHINE
    if reg_path_l == "HKEY_CURRENT_USER":
        reg = winreg.HKEY_CURRENT_USER
    if reg_path_l == "HKEY_LOCAL_MACHINE":
        reg = winreg.HKEY_LOCAL_MACHINE
    if reg_path_l == "HEKY_CLASSES_ROOT":
        reg = winreg.HKEY_CLASSES_ROOT
    if reg_path_l == "HEKY_USERS":
        reg = winreg.HKEY_USERS
    if reg_path_l == "HEKY_CURRENT_CONFIG":
        reg = winreg.HKEY_CURRENT_CONFIG

    return reg


def name_to_const(reg_path):
    reg = winreg.HKEY_LOCAL_MACHINE

    reg_path_l = reg_path.split('\\')[0]

    if reg_path_l == "HKEY_CURRENT_USER":
        reg = winreg.HKEY_CURRENT_USER
    if reg_path_l == "HKEY_LOCAL_MACHINE":
        reg = winreg.HKEY_LOCAL_MACHINE
    if reg_path_l == "HEKY_CLASSES_ROOT":
        reg = winreg.HKEY_CLASSES_ROOT
    if reg_path_l == "HEKY_USERS":
        reg = winreg.HKEY_USERS
    if reg_path_l == "HEKY_CURRENT_CONFIG":
        reg = winreg.HKEY_CURRENT_CONFIG

    return reg


def get_all(reg_path: str):
    """
    sepstr be \\
    """
    d = {}

    reg = name_to_const(reg_path)

    try:
        key = winreg.OpenKey(reg, reg_path)
    except OSError:
        return d

    # 获取该键的所有键值，遍历枚举
    try:
        i = 0
        while 1:
            # EnumValue方法用来枚举键值，EnumKey用来枚举子键
            name, value, _type = winreg.EnumValue(key, i)
            d[name] = (name, value, _type)
            i += 1
    except OSError:
        pass

    return d


def get_keys(reg_path):
    r"""
    sepstr only be '\\'
    """
    reg = name_to_const(reg_path)

    try:
        key = winreg.OpenKey(reg, reg_path)
    except OSError:
        key = winreg.CreateKey(reg, reg_path)

    winreg.CloseKey(key)


classInfo = type | types.UnionType | tuple[object, ...]


def py_type_to_reg_type(pyType: object, big_type: bool = False):
    if isinstance(pyType, int):  # pyType为int
        if big_type:
            return winreg.REG_QWORD
        else:
            return winreg.REG_DWORD
    elif isinstance(pyType, float):  # pyType = float
        return ValueError("REG types not have 'float'.")
    elif isinstance(pyType, str):
        if big_type:
            return winreg.REG_EXPAND_SZ
        else:
            return winreg.REG_SZ
    else:
        return ValueError(f"pyType '{pyType}' don't turn to 'REG_TYPE'.")


def spilt_reg_path(reg_path: str):
    reg = name_to_const(reg_path)
    path = '\\'.join(reg_path.split('\\')[1::])
    return reg, path


def set_reg_key(reg_path: str, _type: int):
    r"""
    reg_path
        path.name=value
    HKEY_CURRENT_USER\SOFTWARE\Microsoft\Windows\CurrentVersion\Policies\Explorer.DisallowRun=1
    """
    after = reg_path.split('.')
    value = after[1].split('=')[1::]
    value = ''.join(value)
    name = after[1].split('=')[0]
    r, p = spilt_reg_path(reg_path)
    p = spilt_reg_path(after[0])[1]
    k = winreg.OpenKey(r, p, 0, winreg.KEY_SET_VALUE)
    winreg.SetValueEx(k, name, 0, _type, value)



from collections import deque

from .token import Token

from .patterns import TokenPatterns
from .process import _process_tokens, _process_indent, _token_strip, _delete_whitespace, _calc_indent_length



# TODO: 词法分析器,Token新的参数


def _lex(code):
    for token_type, pattern in TokenPatterns:
        match = pattern.match(code)
        if match is None:
            # print(f"{token_type} not match")
            continue
        token = Token(token_type, match)
        return token
    # print(f"Longer Token: {longer_token}")
    # return longer_token


# 词法分析器
class Lexer:
    @staticmethod
    def lex(source_code):
        tokens = deque()
        while source_code:
            token = _lex(source_code)
            if token is None:
                current_code = source_code.split('\n')[0]
                raise SyntaxError(f"Invalid syntax: {current_code}")
            tokens.append(token)
            source_code = source_code[len(token):]
        tokens_lst = list(tokens)
        _token_strip(tokens_lst)
        _process_tokens(tokens_lst)
        tokens_lst = _process_indent(tokens_lst)
        _delete_whitespace(tokens_lst)

        return tokens_lst



import json
from . import abc


class Json(abc.Serializer):
    left_delimiter = b"{["
    right_delimiter = b"}]"

    quotations = b"'\""

    mapping = {
        b'{': b'}',
        b'[': b']',
    }

    def __init__(self):
        self.stack = None
        self.s = None

    def dumps(self, data):
        return json.dumps(data).encode()

    def loads(self, data):
        return json.loads(data)


from ...._hycore.data_structures import Stack
from .token import Token


def _calc_indent_length(indent):
    value = indent
    return value.count('\t') * 4 + value.count(' ')


def _process_tokens(tokens: list[Token]):
    i = 0
    length = len(tokens)
    while i < length:
        length = len(tokens)
        if tokens[i].type == 'NEWLINE':
            if i + 1 < length and tokens[i + 1].type == 'NEWLINE':
                tokens.insert(i + 1, Token(
                    'WS', ''
                ))
                # i += 1
            if i + 1 < length and tokens[i + 1].type == 'WS':
                # if i+2 < length and tokens[i+2].type == 'NEWLINE':
                #     tokens.pop(i)
                #     tokens.pop(i+1)
                #     tokens.pop(i+2)
                #     i += 2
                #     continue
                ws = tokens[i + 1]
                tokens[i + 1] = Token(
                    'INDENT', _calc_indent_length(ws.value),
                    tokens[i].lineno, tokens[i].colno, tokens[i].offset + ws.offset
                )
                i += 1

        i += 1


def _process_indent(tokens: list[Token]) -> list[Token]:
    processed_tokens = []
    indent_stack = Stack([0])  # Use a stack to keep track of indentation levels

    for i, token in enumerate(tokens):
        if token.type == 'INDENT':
            current_indent = token.value
            last_indent = indent_stack.at_top

            if current_indent > last_indent:
                # Increase indentation level
                indent_stack.push(current_indent)
                processed_tokens.append(
                    Token(
                        'INDENT', current_indent - last_indent, token.lineno, token.colno, token.offset))
            elif current_indent < last_indent:
                # Decrease indentation level
                while indent_stack and current_indent < indent_stack[-1]:
                    last_indent = indent_stack.pop()
                    processed_tokens.append(Token('DEDENT', last_indent - current_indent))

                if not indent_stack or current_indent != indent_stack[-1]:
                    raise IndentationError(f"Unexpected dedent at position {i}")
            else:
                # No change in indentation
                continue
        else:
            processed_tokens.append(token)

    # Handle any remaining dedents at the end of the file
    while len(indent_stack) > 1:
        last_indent = indent_stack.pop()
        processed_tokens.append(Token('DEDENT', last_indent - indent_stack[-1]))

    return processed_tokens


def _delete_whitespace(tokens: list[Token]):
    i = 0
    while i < len(tokens):
        if tokens[i].type == 'WS':
            tokens.pop(i)
        else:
            i += 1


def _token_strip(tokens: list[Token]):
    i = 0
    while i < len(tokens):
        if tokens[i].type == 'NEWLINE':
            tokens.pop(i)
        else:
            break
        i += 1
    i = len(tokens) - 1
    while i >= 0:
        if tokens[i].type == 'NEWLINE':
            tokens.pop(i)
        else:
            break
        i -= 1



import jsonpickle
from . import S_Json


class JsonPickle(S_Json.Json):
    def dumps(self, data):
        return jsonpickle.dumps(data).encode()

    def loads(self, data):
        return jsonpickle.loads(data)




import logging

# module end 我们会将模块的很多常量慢慢集成到这里

zlibcon_logger = logging.getLogger(__name__)

# const var = [module id][error_const: 1, flag_const: 0, info_const: 2][code]

# flags
DBX_FUNC_TYPE_MRO = 0x1001
DBX_FUNC_TYPE_GET = 0x1002

# flags
ETY_TYPE_KEY = 0x1003
ETY_TYPE_IV = 0x1004

# error
BCR_EXIST_NAME = 0x3101
BCR_NAME_NOT_FOUND = 0x3102
BCR_CANNOT_FOUND = 0x3103
BCR_BAD_REQUEST = 0x3104

# flags
BCR_PERSON = 0x3004
BCR_GROUP = 0x3005
BCR_ALL = 0x3006
BCR_RANGE = 0x3007

BCR_SEND = 0x3008
BCR_OPERATOR = 0x3009

# operation flags
BCR_DISCONNECT = 0x3001

PERM_ROLE_ALLOW = 0x4000
PERM_ROLE_DENY = 0x4001

JSON_DECODE = 0x5001
JSON_ENCODE = 0x5002

UNKNOWN = "Unknown"  # ???

_link_constants = []


def link_constant(name, value):
    """
    create new constant in `libcon`


    :param name:
    :param value:
    :return:
    """
    globals()[name] = value
    _link_constants.append(name)


def unlink_constant(name):
    if name not in _link_constants:
        raise ValueError(f"{name} is not a constant")
    del globals()[name]


zlibcon_logger.debug("Module zlibcon loading ...")



import fnmatch
import os

from .._hycore.json import Pickle
from .._hycore.json_file import json_safe_open, json_write


class CannotFindError(BaseException):
    ...


class MultipleTargetItemsError(BaseException):
    ...


class NotExistError(BaseException):
    ...


class ExistItemError(BaseException):
    ...


errors = CannotFindError, MultipleTargetItemsError, NotExistError, ExistItemError


class Template:
    """
    用于设置数据库的数据模版
    """

    def __init__(self, **dic):
        self._template = dic

    @property
    def template(self):
        """
        返回模版字典
        :return:
        """
        return self._template

    @staticmethod
    def instance(tem, dic):
        """
        判断dic是否符合模版tem
        :param tem:
        :param dic:
        :return:
        """
        return (tem.template.keys()) == (dic.keys())

    def instances(self, dic):
        """
        返回dic是否符合模版
        :param dic:
        :return:
        """
        return self.instance(self, dic)


class DB:
    _template = Template()

    def __init__(self, filename):
        assert fnmatch.fnmatch(filename, "*.db"), "文件名必须以.db结尾"
        self._usetype = list
        self._filename = filename
        self._stream = json_safe_open(filename, dict)
        self._version = self._stream["version"]
        self._name = self._stream['name']
        self._template = self._stream['template']

        self._stream['data'] = self._stream.top('data', {})
        self._stream.indent = 4

    def mkmro(self, mro, template: Template = _template):
        """
        创建一个数据表
        :param mro:
        :param template: 模版对象
        :return:
        """
        if mro in self._stream['data']:
            raise KeyError()
        self._stream['data'][mro] = {'data': self._usetype(), 'template': template}

    def clear(self):
        """
        删除所有的表
        :return:
        """
        return self._stream['data'].clear()

    def exist(self, mro):
        """
        如果表mro存在，返回True，否则False
        :param mro:
        :return:
        """
        return mro in self._stream['data']

    def getfuncOf(self, mro):
        """
        返回一个表的操作对象
        :param mro:
        :return:
        """
        return MroFunc(self, mro)

    def mkget(self, mro, template: Template = _template):
        """
        创建一个表并返回它的操作对象
        :param mro:
        :param template: 模版对象
        :return:
        """
        self.mkmro(mro, template)
        return self.getfuncOf(mro)

    def save(self):
        """
        保存数据库的改动
        :return:
        """
        self._stream.save()

    def __str__(self):
        return f"DBObject \n{Pickle.encode(self._stream['data'])}"

    def __call__(self, _type):
        """
        这个功能用于作为MroFunc的内部接口，请不要随意修改
        :return: Unkown
        """
        if _type == "template":
            return self._template
        if _type == "stream":
            return self._stream
        if _type == "version":
            return self._version
        if _type == "name":
            return self._name


class MroFunc:
    def __init__(self, father: DB, mro):
        self._father = father
        self._mro = mro
        self._stream = self._father("stream")

    @property
    def mro(self):
        return self._mro

    @property
    def father(self):
        return str(self._father)

    @property
    def template(self):
        return self._stream['data'][self.mro]['template']

    def inner(self, dic1: dict, dic2: dict):
        """
        返回dic1是否是dic2的子字典（包括值）
        :param dic1:
        :param dic2:
        :return: bool
        """
        if tuple(dic1.keys()) <= tuple(dic2.keys()):
            for k in dic1:
                if dic1[k] != dic2[k]:
                    return False
            return True
        return False

    def exist(self, **dic):
        return [i for i in self._stream['data'][self.mro]['data'] if self.inner(dic, i)]

    def where(self, **dic):
        """
        找到一个或多个匹配的项，并返回（不存在则返回None）
        :param dic:
        :return:
        """
        ls = []
        for item in self._stream['data'][self.mro]['data']:
            if self.inner(dic, item):
                ls.append(item)
        return ls

    def absquery(self, **dic) -> dict:
        """
        方法要求dic必须有效或是有效数据的一部分，返回唯一的匹配项
        :param dic: 必须有效
        :return:
        """
        ls = self.where(**dic)
        ls_len = len(ls)
        if ls_len < 1:
            raise CannotFindError(f"Cannot find item: {dic}")
        if ls_len > 1:
            raise MultipleTargetItemsError(f"Multiple target items: {ls_len}(except 1), query_data={dic}")
        return ls[0]

    def remove(self, **dic):
        """
        删除一个有效数据，成功True，失败False
        :param dic:
        :return:
        """
        if not self.exist(**dic):
            raise NotExistError(f"Not exist: {dic}")
        try:
            data = self.absquery(**dic)
            self._stream['data'][self.mro]['data'].remove(data)
            return True
        except CannotFindError:

            return False
        except MultipleTargetItemsError:

            return False

    def calibration(self, tem, dic):
        """
        将字典dic转换成完全符合tem的格式
        :param tem: 模版
        :param dic: 字典
        :return:
        """
        for i in tuple(dic):
            if i not in tem.template:
                dic.pop(i)
        for i in tem.template:
            if i not in dic:
                dic[i] = tem.template[i]
        return dic

    def add(self, **dic):
        """
        向表中添加一项数据
        :param dic:数据
        :return:
        """
        # 检查是否符合模版
        dic = self.calibration(self._stream['data'][self.mro]['template'], dic)
        self._stream['data'][self.mro]['data'].append(dic)

    def format(self):
        """
        对每一个有效数据进行检查并修改（可能会丢失一些数据）
        :return:
        """
        mro = self._stream['data'][self.mro]['data']
        tem = self._stream['data'][self.mro]['template']
        for i in range(len(mro)):
            if not Template.instance(tem, mro[i]):
                mro[i] = self.calibration(tem, mro[i])
        self._stream['data'][self.mro]['data'] = mro

    def settemplate(self, template, update=True):
        """
        设置表数据模版，如果update为True，那么调用format方法更新数据
        :param template: 模版对象
        :param update:是否更新
        :return:
        """
        self._stream['data'][self.mro]['template'] = template
        if update:
            self.format()


def setversion(file, version):
    """
    设置数据库的版本(version)
    :param file: 文件路径
    :param version: 版本
    :return:
    """
    f = json_safe_open(file)
    f['version'] = version
    f.close()


def setname(file, name):
    """
    设置数据的内置名称
    :param file: 文件路径
    :param name:数据库命名
    :return:
    """
    f = json_safe_open(file)
    f['name'] = name
    f.close()


def settemplate(file, template):
    """
    设置数据库的默认模版
    :param file:文件路径
    :param template: 模版对象
    :return:
    """
    f = json_safe_open(file)
    f['template'] = template
    f.close()


def mkdb(file):
    """
    创建/清空一个数据库文件
    :param file: 文件路径
    :return:
    """
    mkdbEx(file, Template(), 0, "Unkown")


def mkdbEx(file, template, version, name):
    """
    使用更多选项创建数据库
    :param file: 文件路径
    :param template: 模版对象
    :param version: 版本
    :param name: 数据库命名
    :return:
    """
    json_write({
        "template": template,
        "version": version,
        "name": name,
        "data": {}
    }, file)


def mkget(file):
    """
    创建一个数据库并返回它的操作对象
    :param file: 文件路径
    :return:
    """
    mkdb(file)
    return DB(file)


def mkgetEx(file, template, version, name):
    """
    使用更多选项创建一个数据库并返回它的操作对象
    :param file: 文件路径
    :param template: 模版对象
    :param version: 版本
    :param name: 数据库命名
    :return:
    """
    mkdbEx(file, template, version, name)
    return DB(file)


def getdb(file):
    """
    获取一个数据库的操作对象
    :param file:
    :return:
    """
    if not fnmatch.fnmatch(file, "*.db"):
        raise ValueError("数据库文件名必须以.db结尾")
    return DB(file)


def findDB(path, pattern="*.db"):
    """
    查找path内的所有库文件（*.db）
    :param path: 路径
    :param pattern: 通配符
    :return:
    """
    dbs = []
    for i in os.listdir(path):
        if os.path.isfile(i):
            if fnmatch.fnmatch(i, pattern):
                dbs.append(i)
    return dbs


def checkDB(file):
    """
    检查文件是否是一个数据库文件
    :param file: 文件路径
    :return:
    """
    f = json_safe_open(file)
    res = set(f.keys()) == {'name', 'version', 'data', 'template'}
    return res



import logging

func_logger = logging.getLogger("SzQlib.zencrypt.func")


# module end


def _pkcs7_pad(data: bytes, block_size: int) -> bytes:
    if len(data) >= block_size:
        return data
    padding_length = block_size - len(data) % block_size
    padding = bytes([padding_length]) * padding_length
    return data + padding


def _pkcs7_unpad(padded_data: bytes, block_size: int = 16) -> bytes:
    if not (len(padded_data) % block_size == 0 and 1 <= padded_data[-1] <= block_size):
        return padded_data  # 不需要解填充，直接返回原字节串

    last_byte = padded_data[-1]
    if last_byte > block_size:
        func_logger.error("Invalid padding byte value", last_byte)
        raise ValueError("Invalid padding byte value", last_byte)

    padding_length = last_byte
    if padded_data[-padding_length:] != bytes([last_byte]) * padding_length:
        func_logger.error("Invalid padding pattern")
        raise ValueError("Invalid padding pattern")

    return padded_data[:-padding_length]


def split(text, size=16):
    if len(text) % size == 0:
        return [text[i:i + size] for i in range(0, len(text), size)]
    res_lenght = ((len(text) // size) + 1) * size
    add_lenght = res_lenght - len(text)
    text = pad(text, res_lenght)
    ls = split(text, size)
    ls[-1] = ls[-1][:add_lenght]
    # print(ls)
    return ls


def pad(text: bytes, size=16):
    return _pkcs7_pad(text, size)


def unpad(text: bytes, size=16):
    return _pkcs7_unpad(text, size)



from types import FunctionType

from . import Serializer


class Wrapping(Serializer):
    def __init__(self, dumps: FunctionType, loads: FunctionType):
        self._dumps = dumps
        self._loads = loads

    def dumps(self, obj):
        return self._dumps(obj)

    def loads(self, obj):
        return self._loads(obj)



from abc import ABC
from copy import deepcopy
from types import FunctionType


class Auto(ABC):
    pass


class AutoRegDict(Auto):
    """
    自动注册未存在于字典的键，同时访问时返回默认值

    可以通过指定`isdeepcopy`属性来说明是否对默认值进行copy操作
    """
    default_value = None
    isdeepcopy = True

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def pop(self, key):
        return self._dict.pop(key)

    def copy(self):
        return self._dict.copy()

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()

    def clear(self):
        return self._dict.clear()

    def __contains__(self, key):
        return self._dict.__contains__(key)

    def __len__(self):
        return self._dict.__len__()

    def __getitem__(self, key):
        if key not in self._dict:
            if self.isdeepcopy:
                v = deepcopy(self.default_value)
            else:
                v = self.default_value
            self._dict[key] = v
            return v
        return self._dict[key]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]


class AutoCompare(Auto):
    """
    自动完成比较操作
    通过指定`__compare_attrs__`属性来指定比较的属性，默认为None，
    如果`__compare_attrs__`为None，那么自动比较将不会生效，而是根据比较符返回一个默认值
    如果被比较的对象不是 `AutoCompare` 的实例，那么比较时会按比较列表的第一个属性作为比较属性
    """
    __compare_attrs__ = ()
    __cmp_funcs__ = {
        'eq': lambda x, y: x == y,
        'ne': lambda x, y: x != y,
        'lt': lambda x, y: x < y,
        'gt': lambda x, y: x > y,
        'le': lambda x, y: x <= y,
        'ge': lambda x, y: x >= y
    }

    def _auto_compare_attrs(self, opt, other, defautl=False):
        if opt not in self.__cmp_funcs__:
            return defautl

        func = self.__cmp_funcs__[opt]

        if not isinstance(other, AutoCompare):
            if self.__compare_attrs__:
                value = getattr(self, self.__compare_attrs__[0])
                return func(value, other)

        if self.__compare_attrs__ is None or other.__compare_attrs__ is None:
            return defautl

        my_attr_values = (
            getattr(self, attr) for attr in self.__compare_attrs__)
        other_attr_values = (
            getattr(other, attr) for attr in other.__compare_attrs__)
        return func(my_attr_values, other_attr_values)

    def __eq__(self, other):
        return self._auto_compare_attrs('eq', other, False)

    def __ne__(self, other):
        return self._auto_compare_attrs('ne', other, True)

    def __lt__(self, other):
        return self._auto_compare_attrs('lt', other, False)

    def __gt__(self, other):
        return self._auto_compare_attrs('gt', other, False)

    def __le__(self, other):
        return self._auto_compare_attrs('le', other, True)

    def __ge__(self, other):
        return self._auto_compare_attrs('ge', other, True)


class AutoState(Auto):
    """
    自动完成对象的状态导出和恢复

    如果`_state_attrs`为None，那么将导出所有属性，恢复时同上
    """
    __state_attrs__ = ()

    def __getstate__(self):
        if self.__state_attrs__ is None:
            return self.__dict__
        return {attr: getattr(self, attr) for attr in self.__state_attrs__}

    def __setstate__(self, state):
        if self.__state_attrs__ is None:
            self.__dict__ = state
        else:
            for attr in self.__state_attrs__:
                setattr(self, attr, state[attr])


class AutoOperator(Auto):
    __operator_funcs__ = {}

    def __add__(self, other):
        return self.__operator_funcs__['+'](self, other)

    def __sub__(self, other):
        return self.__operator_funcs__['-'](self, other)

    def __mul__(self, other):
        return self.__operator_funcs__['*'](self, other)

    def __truediv__(self, other):
        return self.__operator_funcs__['/'](self, other)

    def __floordiv__(self, other):
        return self.__operator_funcs__['//'](self, other)

    def __mod__(self, other):
        return self.__operator_funcs__['%'](self, other)

    def __pow__(self, other):
        return self.__operator_funcs__['**'](self, other)

    def __and__(self, other):
        return self.__operator_funcs__['&'](self, other)

    def __or__(self, other):
        return self.__operator_funcs__['|'](self, other)

    def __xor__(self, other):
        return self.__operator_funcs__['^'](self, other)

    def __lshift__(self, other):
        return self.__operator_funcs__['<<'](self, other)

    def __rshift__(self, other):
        return self.__operator_funcs__['>>'](self, other)

    def __invert__(self):
        return self.__operator_funcs__['~'](self)

    def __neg__(self):
        return self.__operator_funcs__['-'](self)


class AutoOperatorStruct:
    @classmethod
    def decorator(cls, oper, globs):
        def wrapper(func):
            return cls(oper, func, globs)

        return wrapper

    def __init__(self, oper: str, func: FunctionType, _globals=None):
        self.oper = oper
        self.func = func

        if oper not in {
            '+', '-', '*', '/', '//', '%', '**', '&', '|', '^', '<<', '>>'
        }:
            raise ValueError(f"{oper} is not a valid operator")

        from ..type_func import get_attr_by_path
        cls = get_attr_by_path(func.__qualname__)
        if not isinstance(cls, AutoOperator):
            raise TypeError(f"{cls} is not a AutoOperator")

        cls.__operator_funcs__[oper] = self

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class AutoOperatorFunc(Auto):
    def __init__(self, globs):
        self._globs = globs

    def operator(self, oper):
        return AutoOperatorStruct.decorator(oper, self._globs)


class AutoRegisterChildMeta(type):
    __children_types__ = []

    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)
        new_cls.__children_types__.append(new_cls)
        return new_cls


class _AutoInfo(Auto):
    ...


class AutoRepr(_AutoInfo):
    __repr_attrs__ = ()

    def __repr__(self):
        return str(
            {attr: getattr(self, attr) for attr in self.__repr_attrs__}
        )


class AutoStr(_AutoInfo):
    _str_attrs = ()

    def __str__(self):
        return str(
            {attr: getattr(self, attr) for attr in self._str_attrs}
        )


class AutoInfo(AutoRepr, AutoStr):
    _info_attrs = ()

    def __repr__(self):
        self._repr_attrs = self._info_attrs
        return super().__repr__()

    def __str__(self):
        self._str_attrs = self._info_attrs
        return super().__str__()



class ExitFunction:
    def __init__(self, func, embedding=True, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

        self.embedding = embedding

        self.name = None

    def __call__(self):
        if self.embedding:
            self.func(*self.args, **self.kwargs)

    def __repr__(self):
        return f"ExitFunction({self.func.__name__})"


class ExitContainer:
    """
    用法类似于 .hyconfig 模块
    class MyExitContainer(ExitContainer):
        exit_1 = ExitFunction(<some function>, True, <args>, <kwargs>)
        exit_2 = ExitFunction(<some function>, False, <args>, <kwargs>)
        ...

    if __name__ == '__main__':
        # No args should you give
        exit_container = MyExitContainer()
        
        # You can get all exit functions
        print(exit_container.exit_functions)  
        
        # The `embedding` argument is used to determine whether to call the function when the parser exits.
        print(exit_container.embedding("exit_1"))  # True
        
        # If you want to call the function when the parser exits, set it to True.
        exit_container.embedding("exit_2", True)
    
        # Or you can set it to False.
        exit_container.embedding("exit_1", False)
        
        # Your code
        # ...
        
        # When the parser exits, all ExitFunction will be called automatically.

    """
    @property
    def exit_functions(self):
        for _ in dir(self):
            value = getattr(self, _)
            if isinstance(value, ExitFunction):
                yield _, value

    def _build_mapping(self):
        for name, value in self.exit_functions:
            self.__mapping[name] = value

    def _add_atexit(self):
        import atexit
        atexit.register(self._at_exit)

    def _at_exit(self):
        for _, value in self.exit_functions:
            value()

    def __init__(self):
        self.__mapping = {}  # type: dict[str, ExitFunction]
        self._build_mapping()
        self._add_atexit()
    
    def embedding(self, name, value=None):
        if name not in self.__mapping:
            raise KeyError(f"{name} is not in {self.__mapping}")
        
        if value is None:
            return self.__mapping[name].embedding

        self.__mapping[name].embedding = value
    
    def origin_function(self, name, value=None):
        if name not in self.__mapping:
            raise KeyError(f"{name} is not in {self.__mapping}")
        
        if value is None:
            return self.__mapping[name].func
        
        self.__mapping[name].func = value





import re
from typing import Union, Any


class Token:
    def __init__(
            self,
            type_: str, value: Union[re.Match, Any],
            lineno: int = 0, colno: int = 0, offset: int = 0
    ):
        self.type = type_
        self.lineno = lineno
        self.colno = colno
        self.offset = offset

        if isinstance(value, re.Match):
            self.match = value
            self.value: Union[str, int] = value.group()
        else:
            self.value: Union[str, int] = value

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.type == other.type and self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, tuple):
            return self.type == other[0] and self.value == other[1]
        else:
            return False

    @property
    def t(self):
        """
        获取Token的类型
        """
        return self.type

    @property
    def v(self):
        """
        获取Token的值
        """
        return self.value

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return repr(self)

    def __len__(self):
        return len(str(self.value))

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.type},\t{repr(self.value)},\t{self.lineno}:{self.colno}-{self.offset}>"



from pyparsing import (
    Word, alphas, alphanums, oneOf, ZeroOrMore, Group, LineEnd, LineStart,
    Optional, Literal, restOfLine
)

# 定义基础元素
identifier = Word(alphas, alphanums + "_")
integer = Word("0123456789")
keyword = oneOf("import from as name")
newline = LineEnd().suppress()
indent = LineStart() + (" " * 4)
dedent = LineStart() - (" " * 4)

# 定义表
table_def = Group(
    Literal("[") + identifier + Literal("]") + newline
)

# 定义变量赋值
assignment = Group(
    identifier + Literal("=") + restOfLine + newline
)

# 定义模块导入
import_statement = Group(
    keyword("import") + identifier("module") + Optional(keyword("as") + identifier("alias")) + newline
)

# 定义从模块导入
from_import_statement = Group(
    keyword("from") + identifier("module") + keyword("import") + identifier("name") +
    Optional(
        keyword("as") + identifier("alias")) + newline
)

# 定义整个代码块
code_block = ZeroOrMore(
    table_def
    | assignment
    | import_statement
    | from_import_statement
)


def parse(code):
    return code_block.parseString(code)



def singleton_decorator(cls):
    class wrap(cls):
        _Instance = None

        def __new__(cls, *args, **kwargs):
            if wrap._Instance is None:
                wrap._Instance = super().__new__(cls, *args, **kwargs)
            return wrap._Instance

        def __str__(self):
            return super().__str__()

        def __repr__(self):
            return super().__repr__()

    return wrap


def Instance(*args, **kwargs):
    def wrapper(cls):
        return cls(*args, **kwargs)

    return wrapper



import base64
import hashlib
import hmac
import json
import logging
import os
from abc import ABC, abstractmethod
from typing import Union, Optional

from .._hycore import type_func, hash
from .._hydatabase import database

zauth_logger = logging.getLogger(__name__)


class SimpleUserPasswdManager:
    def __init__(self, db_file, salt_lenght=64):
        self._salt_lenght = salt_lenght
        try:
            self._db = database.DB(db_file)
        except KeyError:
            self._db = database.mkget(db_file)
        self._template = type_func.Template(
            username="Unknown",
            hash_type="sha256",
            hash_lenght=64,  # set this param if you use shake_128 or shake_256
            passwd_salt=None,
            passwd_hash=None,
        )
        if not self._db.exist("UserCredentials"):
            self._db.mkmro("UserCredentials", template=self._template)

        self._mrofunc = self._db.getfuncOf("UserCredentials")

    def add(self, username: str, passwd: str, hash_type="sha256", hash_lenght=64):
        if self._mrofunc.exist(username=username):
            raise database.ExistItemError(username)

        salt = os.urandom(self._salt_lenght)
        passwd_salt = passwd.encode() + salt
        passwd_hash = hash.getHashValueByName(passwd_salt, hash_type, hash_lenght)

        self._mrofunc.add(
            username=username,
            hash_type=hash_type,
            hash_lenght=hash_lenght,
            passwd_salt=salt,
            passwd_hash=passwd_hash
        )

    def delete(self, username):
        try:
            usr = self._mrofunc.absquery(username=username)
            self._mrofunc.remove(**usr)
        except AssertionError:
            return

    def exist(self, username):
        return self._mrofunc.exist(username=username)

    def query(self, username, passwd):
        try:
            usr = self._mrofunc.absquery(username=username)
            salt = usr["passwd_salt"]
            user_passwd_bytes = passwd.encode()

            true_passwd_hash = usr["passwd_hash"]

            user_passwd_salt = user_passwd_bytes + salt

            if hash.getHashValueByName(user_passwd_salt, usr["hash_type"], usr["hash_lenght"]) == true_passwd_hash:
                return True
            else:
                return False

        except database.errors:
            return


# 作为所有Token类的基类

class TokenBase(ABC):
    @abstractmethod
    def serialize(self):
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, *args, **kwargs):
        pass


class JsonWebToken(TokenBase):
    def __init__(self, head: dict, payload: dict, secret_key: str):
        self.head = head
        self.payload = payload
        self.secret = secret_key
        self.signature = self._generate_signature()

        self._head = None
        self._payload = None

    def flush(self):
        header = base64.urlsafe_b64encode(json.dumps(self.head).encode()).decode().rstrip('=')
        payloader = base64.urlsafe_b64encode(json.dumps(self.payload).encode()).decode().rstrip('=')
        self._head, self._payload = header, payloader
        return header, payloader

    def _generate_signature(self):
        # 将头部和负载部分进行 base64url 编码
        # 拼接头部和负载部分
        message = f"{self._head}.{self._payload}".encode()

        # 计算 HMAC-SHA256 签名
        signature = hmac.new(self.secret.encode(), msg=message, digestmod=hash.hashlib.sha256).digest()

        # 对签名进行 base64url 编码
        return base64.urlsafe_b64encode(signature).decode().rstrip('=')

    def __str__(self):
        self.flush()
        signature = self.signature

        return f"{'.'.join((self._head, self._payload))}.{signature}"

    def serialize(self) -> str:
        self.flush()
        signature = self.signature

        return f"{'.'.join((self._head, self._payload))}.{signature}"

    @classmethod
    def deserialize(cls, secret, token_str: str) -> 'JsonWebToken':
        parts = token_str.split('.')
        if len(parts) != 3:
            raise ValueError("Invalid token format")

        head = json.loads(base64.urlsafe_b64decode(parts[0] + '==').decode())
        payload = json.loads(base64.urlsafe_b64decode(parts[1] + '==').decode())
        signature = base64.urlsafe_b64decode(parts[2] + '==')

        # 验证签名
        expected_signature = hmac.new(
            secret.encode(),
            msg=f"{base64.urlsafe_b64encode(json.dumps(head).encode()).decode()}."
                f"{base64.urlsafe_b64encode(json.dumps(payload).encode()).decode()}".encode(),
            digestmod=hashlib.sha256).digest()
        if not hmac.compare_digest(signature, expected_signature):
            raise ValueError("Invalid signature")

        return cls(head, payload, secret)


class OAuthAccessToken(TokenBase):
    def __init__(self, token: str, expires_in: int, scope: str):
        self.token = token
        self.expires_in = expires_in
        self.scope = scope

    def serialize(self):
        return {
            "token": self.token,
            "expires_in": self.expires_in,
            "scope": self.scope
        }

    @classmethod
    def deserialize(cls, token_str):
        data = json.loads(token_str)
        return cls(data["token"], data["expires_in"], data["scope"])


class RefreshToken(TokenBase):
    def __init__(self, token: str, expires_in: int):
        self.token = token
        self.expires_in = expires_in

    def serialize(self):
        return {
            "token": self.token,
            "expires_in": self.expires_in
        }

    @classmethod
    def deserialize(cls, token_str):
        data = json.loads(token_str)
        return cls(data["token"], data["expires_in"])


class SessionToken(TokenBase):
    def __init__(self, session_id: str, user_id: str, expires_at: int):
        self.session_id = session_id
        self.user_id = user_id
        self.expires_at = expires_at

    def serialize(self):
        return {
            "session_id": self.session_id,
            "user_id": self.user_id,
            "expires_at": self.expires_at
        }

    @classmethod
    def deserialize(cls, token_str):
        data = json.loads(token_str)
        return cls(data["session_id"], data["user_id"], data["expires_at"])


class HashToken(TokenBase):
    def __init__(self, original_token: Optional[bytes] = None, salt: Optional[bytes] = None):
        if original_token is None:
            original_token = os.urandom(32)

        if salt is None:
            salt = os.urandom(16)

        self.original_token = original_token
        self.salt = salt
        self.token = None

    def serialize(self):
        return hash.getHashValueByName(self.original_token + self.salt, "sha256")

    @classmethod
    def deserialize(cls, *args, **kwargs):
        raise Exception("无法反序列化基于不可逆运算的令牌。")

    def __eq__(self, other: 'HashToken'):
        return self.serialize() == other.serialize()


class HashTokenManager:
    class _HashTokenComparator(HashToken):
        """
        内置的令牌比较器，用于字节串与令牌对象的比较
        """

        def __init__(self, x):
            super().__init__()
            self.token_cmp = x

        def serialize(self):
            return self.token_cmp

        def __eq__(self, other):
            return self.serialize() == other.serialize()

    def __init__(self):
        self._tokens = set()

    def add(self, token: HashToken):
        """
        添加一个令牌对象
        """
        if token in self._tokens:
            raise Exception("Token already exists.")
        else:
            self._tokens.add(token)
            return True

    def add_token(self, original_token: Optional[bytes] = None, salt: Optional[bytes] = None):
        """
        创建一个令牌对象，并添加进内部数据库中
        :param original_token: 原始令牌数据
        :param salt: 随机盐
        """
        obj = HashToken(original_token, salt)
        self.add(obj)
        return obj

    TokenValueType = Union[bytes, HashToken, _HashTokenComparator]

    def query(self, token: 'TokenValueType'):
        """
        查询一个令牌
        :param token: 查询的令牌，可以是字节串，令牌对象中的一种
        """
        if isinstance(token, HashToken):
            return token in self._tokens
        if isinstance(token, bytes):
            return self.query(self._HashTokenComparator(token))

    def remove(self, token: 'TokenValueType'):
        """
        删除一个令牌
        :param token: 删除的令牌，可以是字节串、令牌对象中的一种
        """
        if isinstance(token, HashToken):
            if token in self._tokens:
                self._tokens.remove(token)
                return token
            else:
                return False
        if isinstance(token, bytes):
            return self.remove(self._HashTokenComparator(token))


zauth_logger.debug("Module Auth loading ...")



import json


def encode(data, indent=None):
    return json.dumps(data, ensure_ascii=False, sort_keys=True, indent=indent)


def decode(data):
    return json.loads(data)


def dump(data, file, *args, **kwargs):
    return json.dump(data, file, *args, **kwargs)


def load(file, *args, **kwargs):
    return json.load(file, *args, **kwargs)



class UndirectedGraph:
    def __init__(self, graph_dict=None):
        """ initializes a graph object
            If no dictionary or None is given,
            an empty dictionary will be used
        """
        if graph_dict is None:
            graph_dict = {}
        self.graph_dict = graph_dict

    def vertices(self):
        """ returns the vertices of a graph """
        return list(self.graph_dict.keys())

    def edges(self):
        """ returns the edges of a graph """
        return self.__generate_edges()

    def add_vertex(self, vertex):
        """ If the vertex "vertex" is not in
            self.__graph_dict, a key "vertex" with an empty
            list as a value is added to the dictionary.
            Otherwise, nothing has to be done.
        """
        if vertex not in self.graph_dict:
            self.graph_dict[vertex] = []
            return True
        return False

    def add_edge(self, left, right):
        """ assumes that edge is of type set, tuple or list;
            between two vertices can be multiple edges!
        """
        (vertex1, vertex2) = left, right

        # Add vertex1 if it does not exist
        if vertex1 not in self.graph_dict:
            self.graph_dict[vertex1] = []

        # Add vertex2 if it does not exist
        if vertex2 not in self.graph_dict:
            self.graph_dict[vertex2] = []

        # Add vertex2 to vertex1's adjacency list if not already present
        if vertex2 not in self.graph_dict[vertex1]:
            self.graph_dict[vertex1].append(vertex2)

        # Add vertex1 to vertex2's adjacency list if not already present
        if vertex1 not in self.graph_dict[vertex2]:
            self.graph_dict[vertex2].append(vertex1)

    def __generate_edges(self):
        """ A static method generating the edges of the
            graph "graph". Edges are represented as sets
            with one (a loop back to the vertex) or two
            vertices
        """
        edges = []
        for vertex in self.graph_dict:
            for neighbour in self.graph_dict[vertex]:
                if {neighbour, vertex} not in edges:
                    edges.append({vertex, neighbour})
        return edges

    def __str__(self):
        res = "vertices: "
        for k in self.graph_dict:
            res += str(k) + " "
        res += "\nedges: "
        for edge in self.__generate_edges():
            res += str(edge) + " "
        return res


class DirectedGraph:
    def __init__(self, graph_dict=None):
        """ initializes a directed graph object
            If no dictionary or None is given,
            an empty dictionary will be used
        """
        if graph_dict is None:
            graph_dict = {}
        self.graph_dict = graph_dict

    def vertices(self):
        """ returns the vertices of a graph """
        return list(self.graph_dict.keys())

    def edges(self):
        """ returns the edges of a graph """
        return self.__generate_edges()

    def add_vertex(self, vertex):
        """ If the vertex "vertex" is not in
            self.graph_dict, a key "vertex" with an empty
            list as a value is added to the dictionary.
        """
        if vertex not in self.graph_dict:
            self.graph_dict[vertex] = []

    def add_edge(self, left, right):
        """ assumes that edge is of type tuple (vertex1, vertex2);
            adds a directed edge from vertex1 to vertex2.
        """
        (vertex1, vertex2) = left, right
        if vertex1 in self.graph_dict:
            self.graph_dict[vertex1].append(vertex2)
        else:
            self.graph_dict[vertex1] = [vertex2]

    def __generate_edges(self):
        """ A static method generating the edges of the
            directed graph "graph". Edges are represented as lists
            with two vertices
        """
        edges = []
        for vertex in self.graph_dict:
            for neighbour in self.graph_dict[vertex]:
                if [vertex, neighbour] not in edges:
                    edges.append([vertex, neighbour])
        return edges

    def __str__(self):
        res = "vertices: "
        for k in self.graph_dict:
            res += str(k) + " "
        res += "\nedges: "
        for edge in self.__generate_edges():
            res += str(edge) + " "
        return res


class WeightedGraph:
    def __init__(self):
        self.graph_dict = {}

    def add_vertex(self, vertex):
        """ Adds a new vertex to the graph """
        if vertex not in self.graph_dict:
            self.graph_dict[vertex] = {}

    def add_edge(self, vertex1, vertex2, weight):
        """ Adds a weighted edge between vertex1 and vertex2 """
        if vertex1 in self.graph_dict and vertex2 in self.graph_dict:
            self.graph_dict[vertex1][vertex2] = weight
            self.graph_dict[vertex2][vertex1] = weight  # For undirected graph

    def get_weight(self, vertex1, vertex2):
        """ Returns the weight of the edge between vertex1 and vertex2 """
        if vertex2 in self.graph_dict[vertex1]:
            return self.graph_dict[vertex1][vertex2]
        return None

    def __str__(self):
        res = "vertices: "
        for k in self.graph_dict:
            res += str(k) + " "
        res += "\nedges: "
        for vertex in self.graph_dict:
            for neighbour in self.graph_dict[vertex]:
                res += f"{vertex} -> {neighbour}, weight={self.graph_dict[vertex][neighbour]} "
        return res


class LabeledGraph:
    def __init__(self):
        self.vertices = {}
        self.edges = {}

    def add_vertex(self, vertex, label=None):
        """ Adds a new vertex with an optional label """
        self.vertices[vertex] = label

    def add_edge(self, vertex1, vertex2, label=None):
        """ Adds an edge between vertex1 and vertex2 with an optional label """
        if vertex1 in self.vertices and vertex2 in self.vertices:
            self.edges[(vertex1, vertex2)] = label

    def get_vertex_label(self, vertex):
        """ Returns the label of a vertex """
        return self.vertices.get(vertex)

    def get_edge_label(self, vertex1, vertex2):
        """ Returns the label of the edge between vertex1 and vertex2 """
        return self.edges.get((vertex1, vertex2))

    def __str__(self):
        res = "vertices: "
        for vertex, label in self.vertices.items():
            res += f"{vertex}({label}) "
        res += "\nedges: "
        for (vertex1, vertex2), label in self.edges.items():
            res += f"{vertex1} -> {vertex2}({label}) "
        return res



class Node:
    def __init__(self, data=None, *children, parent=None):
        self.data = data
        self.children = list(children)
        self.parent = parent

    def add_child(self, node):
        self.children.append(node)

    def remove_child(self, index):
        self.children.remove(index)

    def index(self, node):
        return self.children.index(node)

    def __str__(self):
        return f"TreeNode: data={self.data}, children_count={len(self.children)}"



import re


class BaseRe:
    pattern = None

    def match(self, string):
        return re.match(self.pattern, string)

    def findall(self, string):
        return re.findall(self.pattern, string)

    def finditer(self, string):
        return re.finditer(self.pattern, string)

    def __str__(self):
        return self.pattern


class Re(BaseRe):
    def __init__(self, pattern):
        self.pattern = pattern


class Literal(BaseRe):
    def __init__(self, literal):
        self.pattern = re.escape(literal)



# from . import connect_length
from . import Json
from .S_BinStruct.methods import connect_length as _connect_length, get_part
from .abc import Serializer
from ..._hycore.type_func import Offset, get_subclasses_recursion, get_type_name


# 找不到序列化器
class CannotFindSerializer(RuntimeError): ...


# 序列化器冲突
class SerializerConflict(RuntimeError): ...


class UnsetSerializer(RuntimeError): ...


def dumps(obj, serializer: Serializer = None):
    """
    序列化数据
    :param obj: 需要序列化的对象
    :param serializer: 指定一个序列化器，如果为 None，则使用 Json
    """
    if serializer is None:
        serialier = Json()

    return \
            _connect_length(
                serializer.dumps(obj)
            ) + \
            _connect_length(
                serializer.__class__.__name__.encode()
            )


def loads(data: bytes, serializer: Serializer = None, ignore=False, use_my_serializer=False):
    """
    从字节串反序列化数据
    :param data: 字节串
    :param serializer: 指定一个序列化器，如果为 None，则使用 Json
    :param ignore: 当指定的序列化器与序列化数据中标记的序列化器不一样时,忽略标记的序列化器，使用指定的序列化器
    :param use_my_serializer: 强制使用指定的序列化器,优先级大于ignore
    :except CannotFindSerializer: 找不到标记的序列化器
    :except SerializerConflict: 指定的序列化器与标记的序列化器不一样,且ignore为False
    :except UnsetSerializer: 未设置序列化器
    """
    offset = Offset(data)

    data = get_part(offset)
    serializer_name = get_part(offset).decode()

    subclasses = {
        t.__name__: t
        for t in get_subclasses_recursion(Serializer)
    }

    if serializer_name not in subclasses:
        # raise ValueError(f'标记的序列化器为 {serializer_name}, 但它不是 Serializer 的子类')
        raise ValueError(f'Got {serializer_name}, but it is not a subclass of Serializer')

    final_serializer = subclasses[serializer_name]()

    if (serializer is not None and
            get_type_name(final_serializer) != get_type_name(serializer)):
        if not ignore:  # 两个序列化器不匹配,且不忽略错误
            # raise ValueError(f'指定的序列化器与标记的序列化器不匹配')
            raise ValueError(f'Specified serializer does not match the marked serializer')

    if use_my_serializer:  # 当use_my_serializer为True时,使用指定的序列化器
        final_serializer = serializer

    if final_serializer is None:
        raise UnsetSerializer()

    return final_serializer.loads(data)



import builtins
import sys
from types import NoneType
from typing import Union

from ...._hycore import type_func


def get_class(equal_name: str):
    cls = type_func.get_attr_by_path(equal_name, {'builtins': builtins, **sys.modules})
    return cls


def get_attr_bitmap_length(number_of_attrs: int):
    return (number_of_attrs + 7) // 8


def length_to_bytes(length_or_obj):
    length = length_or_obj if isinstance(length_or_obj, int) else len(length_or_obj)
    return type_func.Number.pack_variable_length_int(length)


def connect_length(bytes_data: Union[bytes, NoneType], no_none=True):
    if not no_none and bytes_data is None:  # NoneType
        return b''
    return b''.join([
        length_to_bytes(bytes_data), bytes_data
    ])


def get_length_offset(offset):
    length, con = type_func.Number.unpack_variable_length_int(offset.surplus(bytes))
    offset += con
    return length


def get_part(offset) -> bytes:
    length = get_length_offset(offset)
    return offset >> length



class Namespace:
    _init = False

    def __init__(self, error='N/A', **kwargs):
        self._namesp = kwargs
        self.error = error

    def set(self, key, value):
        self._namesp[key] = value

    def get(self, item):
        if item in self._namesp:
            return self._namesp[item]
        else:
            return self.error

    def __getattr__(self, item):
        if item in self._namesp:
            return self._namesp[item]
        else:
            return self.error

    def delt(self, item):
        if item in self._namesp:
            self._namesp.pop(item)

    def set_namespace(self, nsp: dict):
        self._namesp = nsp

    def __setitem__(self, key, value):
        self.set(key, value)

    def __delitem__(self, key):
        self.delt(key)

    def __getitem__(self, item):
        return self.get(item)

    def items(self):
        return self._namesp.items()

    def keys(self):
        return self._namesp.keys()

    def values(self):
        return self._namesp.values()



from pathlib import Path
from typing import Union, Type

from .const import *
from .items import ConfigItem, ConfigItemInstance
from ..abc.backend import BackendABC
from ..._hycore.utils import DoubleDict


def get_keys_by_type(dct, tp):
    ls = set()
    for attr in dct:
        if attr.startswith('__'):
            continue
        value = dct[attr]
        if isinstance(value, tp):
            ls.add(attr)
    return ls


def get_attrs_by_type(obj, tp):
    ls = set()
    for attr in dir(obj):
        if attr.startswith('__'):
            continue
        value = getattr(obj, attr)
        if isinstance(value, tp):
            ls.add(attr)
    return ls


class ConfigError(Exception):
    ...


class NewConfigContainerError(ConfigError):
    def __init__(self, e):
        self.e = e

    def __str__(self):
        return f"Failed to create config container ==> {self.e}"


def _build_items(self: 'type[ConfigContainer]'):
    self.__cfgitems__ = get_attrs_by_type(self, ConfigItem)


def _build_mapping(self: 'type[ConfigContainer]'):
    mapping = self.__cfgmapping__
    items = self.__cfgitems__

    for name in items:
        item = getattr(self, name)  # type: ConfigItem
        key = item.key

        if item.key == AUTO:
            item.key = key = name

        if name in mapping:  # 有配置项的name与其他配置项的key或name冲突
            raise NewConfigContainerError(f'{name} is a config item and key conflict')
        if key in mapping:  # 有配置项的key与其他配置项的key或name冲突
            raise NewConfigContainerError(f'{key} is a config item and key conflict')

        item.attr = name

        mapping[key] = name  # 保存映射关系


class ConfigContainer:
    """
    配置主类
    继承并添加ConfigItem类属性
    所有ConfigContainer是通用的
    比如:
    ```
    class MyConfig(ConfigContainer):
        configItem1 = ConfigItem('configItem1', type=IntType, default=0)
        configItem2 = ConfigItem('configItem2', type=BoolType, default=True)

        # configItemError1 = ConfigItem('configItemError1', type=IntType, default='123')
        #   这将引发TypeError,您应该保证default和type的允许类型是一样的

        configItem3 = ConfigItem('configItem3-key', type=ListType, default=[])
        #   您可以随意指定配置项的键,这个键作为配置文件中显示的键

        # ConfigContainer会自动完成key_to_attr的转换,只要你使用__getitem__和__setitem__方法,注意,这些方法的会将属性名作为转换标准

        # configItem4 = ConfigItem('configItem3', ...)  # Error
        #   您无法将一个配置项的key设置成已存在的配置项的名称或键,这将会在定义时报错: configItem3 is a config item and key conflict
    ```
    """

    # 内部属性
    __cfgitems__: set = None
    __cfgmapping__: DoubleDict = None
    __cfgbackend_instance__: BackendABC = None

    # 可重写配置属性
    __cfgfile__: Union[Path, str] = None
    __cfgbackend__: Type[BackendABC] = None
    __cfgautoload__ = False

    @property
    def cfg_items(self):
        return self.__cfgitems__

    @property
    def cfg_backend_type(self):
        return self.__cfgbackend__

    @property
    def cfg_backend(self):
        return self.__cfgbackend_instance__

    @property
    def cfg_mapping(self):
        return self.__cfgmapping__

    @property
    def cfg_file(self):
        return self.__cfgfile__

    @property
    def cfg_autoload(self):
        return self.__cfgautoload__

    @classmethod
    def get_cfgitem(cls, name, instance) -> 'ConfigItemInstance':
        return getattr(cls, name).from_instance(instance)

    def __init_subclass__(cls, **kwargs):  # 对于每一个子类,都会执行一次映射构建
        cls.__cfgitems__ = cls.__cfgitems__ or set()
        cls.__cfgmapping__ = DoubleDict()

        _build_items(cls)
        _build_mapping(cls)

    def __init__(self):
        self.changes: set = set()

        self.__cfgbackend_instance__ = self.__cfgbackend__()  # 创建后端实例

        if self.cfg_autoload:
            self.load(self.cfg_file)

    def load(self, file=None):
        self.cfg_backend.file = file
        self._load()

    def validate(self, key_or_attr, value, error=False):
        if not self.exists(key_or_attr):
            raise KeyError(f'{key_or_attr} is not a valid config item or key')
        return self.get_cfgitem(key_or_attr, self).validate(value, error)

    @property
    def existing(self):
        """
        加载时配置文件是否存在
        """
        return self.cfg_backend.existing

    def exists(self, key_or_attr):
        """
        判断配置项是否存在
        """
        attr = self.to_attr(key_or_attr)
        return attr in self.config_names()

    def config_names(self):
        """
        返回作为配置项的属性名集合
        """
        return self.cfg_items

    def config_values(self):
        """
        返回作为配置项的属性值集合
        """
        return [getattr(self, key) for key in self.config_names()]

    def config_items(self):
        """
        返回作为配置项的属性名和属性值集合
        """
        return [(key, getattr(self, key)) for key in self.config_names()]

    def save(self):
        """
        保存配置(仅限所有改动)
        """
        self._save(self.changes)

    def save_all(self):
        """
        保存所有配置
        """
        self._save(self.config_names())

    def clear_changes(self):
        self.changes.clear()

    def reset(self):
        for attr in self.config_names():
            data = self.get_cfgitem(attr, self)
            setattr(self, attr, data.default)

    def to_attr(self, key_or_attr):
        if key_or_attr in self.cfg_items:
            return key_or_attr
        elif key_or_attr in self.cfg_mapping:
            return self.cfg_mapping[key_or_attr]
        else:
            raise KeyError(f'{key_or_attr} is not a valid config item or key')

    def on_change(self, key, old, new):
        self.changes.add(key)

    def _load(self):
        self.cfg_backend.load()
        for key, value in self.cfg_backend.items():
            attr = self.to_attr(key)
            data = self.get_cfgitem(attr, self)

            if data.validate(value):
                setattr(self, attr, value)
            else:
                setattr(self, attr, data.default)

        self.changes.clear()

    def __getitem__(self, key):
        attr = self.to_attr(key)
        if attr in self.config_names():
            return getattr(self, attr)
        else:
            raise KeyError(f'{key} is not a config item or key')

    def __setitem__(self, key, value):
        attr = self.to_attr(key)
        if attr in self.config_names():
            setattr(self, attr, value)
        else:
            raise KeyError(f'{key} is not a config item or key')

    @classmethod
    def _get_item(cls, item_attr) -> ConfigItem:
        return getattr(cls, item_attr)

    def _save(self, keys_or_attrs):
        for name in keys_or_attrs:
            attr = self.to_attr(name)
            value = self.get_cfgitem(name, self)
            self.cfg_backend.set(name, value.tp_transform())

        self.cfg_backend.save()




# empty/example module


data = {
    'email': 'idesong6@qq.com',
    'author': 'SongzqInChina',
    'project':{
        'name': 'HydrogenLib',
        'version': 'None'
    },
    'special_chars': "\n\r\a \\ '",
    'special_data': {
        '[': '}',
        '{': ']',
        'd':[
            'a', 'b', 'c'
        ]
    }
}



from typing import Union

from ..utils.auto import AutoRegDict, AutoCompare
from ..utils.namespace import Namespace


class BaseStruct(AutoCompare):
    __compare_attrs__ = ('name',)

    def __init__(self, name):
        self.name = name
        self.joins = set()
        self.info = Infomation()


class BaseError(Exception):
    ...


class ObjectExistError(BaseError):
    ...


class Infomation(Namespace):
    ...


class User(BaseStruct):
    def __init__(self, name):
        super().__init__(name)
        self.info = Infomation()


class Group(BaseStruct):
    def __init__(self, name):
        super().__init__(name)
        self.info = Infomation()


class Domain(BaseStruct):
    def __init__(self, name):
        super().__init__(name)
        self.info = Infomation()


priority_dict = {
    "User": 0,
    "Group": 1,
    "Domain": 2,
    User: 0,
    Group: 1,
    Domain: 2
}


class Manager:
    def __init__(self, users=None, groups=None, domains=None):
        self.objects = AutoRegDict()
        if users is not None:
            for user in users:
                self.add_user(user)
        if groups is not None:
            for group in groups:
                self.add_group(group)

        if domains is not None:
            for domain in domains:
                self.add_domain(domain)

    def add_user(self, user: User) -> bool:
        if user.name in self.objects:
            return False
        self.objects[user.name] = user
        return True

    def remove_user(self, user: User) -> bool:
        if user.name not in self.objects:
            return False
        del self.objects[user.name]
        return True

    def add_group(self, group: Group) -> bool:
        if group.name in self.objects:
            return False
        self.objects[group.name] = group
        return True

    def remove_group(self, group: Group) -> bool:
        if group.name not in self.objects:
            return False
        del self.objects[group.name]
        return True

    def add_domain(self, domain: Domain) -> bool:
        if domain.name in self.objects:
            return False
        self.objects[domain.name] = domain
        return True

    def remove_domain(self, domain: Domain) -> bool:
        if domain.name not in self.objects:
            return False
        del self.objects[domain.name]
        return True

    def create_user(self, name):
        if name in self.objects:
            raise ObjectExistError(f"{name} is already exist")
        u = User(name)
        self.objects[name] = u

    def create_group(self, name):
        if name in self.objects:
            raise ObjectExistError(f"{name} is already exist")
        g = Group(name)
        self.objects[name] = g

    def create_domain(self, name):
        if name in self.objects:
            raise ObjectExistError(f"{name} is already exist")
        d = Domain(name)
        self.objects[name] = d

    _ContainerType = Union[Group, Domain]
    _NonTopType = Union[User, Group]

    def let_join(self, name, container):
        nontop_obj = self.objects[name]
        container_obj = self.objects[container]
        if not isinstance(container_obj, self._ContainerType):
            raise TypeError(f'{container_obj} is not a container')
        if not isinstance(nontop_obj, self._NonTopType):
            raise TypeError(f'{nontop_obj} is not a non-top object')
        if priority_dict[type(nontop_obj)] > priority_dict[type(container_obj)]:
            raise TypeError(f'{nontop_obj} is not a {container_obj}')
        # TODO: 完成容器加入操作， 完成容器离开操作
        if container_obj in nontop_obj.joins:
            raise ObjectExistError(f"{nontop_obj} is already in {container_obj}")
        nontop_obj.joins.add(container_obj)

    def let_leave(self, name, container):
        nontop_obj = self.objects[name]
        container_obj = self.objects[container]
        if not isinstance(container_obj, self._ContainerType):
            raise TypeError(f'{container_obj} is not a container')
        if not isinstance(nontop_obj, self._NonTopType):
            raise TypeError(f'{nontop_obj} is not a non-top object')
        if container_obj not in nontop_obj.joins:
            raise ObjectExistError(f"{nontop_obj} is not in {container_obj}")
        nontop_obj.joins.remove(container_obj)

    def get_info(self, name):
        if name not in self.objects:
            return None
        return self.objects[name].info

    def get_obj(self, name):
        return self.objects[name]



# Namedpipe for windows

# module end
# module start at Aug 31st 2024 20:31
# module not end
# TODO: 完成命名管道的重构


# self.handle = win32pipe.CreateNamedPipe(  Server
#     self.format_name(),
#     win32pipe.PIPE_ACCESS_INBOUND,
#     win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_WAIT,
#     1, 65535, 65535,
#     0, None
# )
# self.handle = win32file.CreateFileW(  Client
#     self.format_name(),
#     win32file.GENERIC_WRITE,
#     0,
#     None,
#     win32file.OPEN_EXISTING,
#     0,
#     None
# )

# Server: Create namedpipe
# Client: Create file
# Server: Connect namedpipe
# Client: Write to file




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'COMMENT DEDENT DYNAMIC_PLACEHOLDER EQUALS IDENTIFIER INDENT LBRACE LBRACKET MULTILINE_COMMENT NUMBER RBRACE RBRACKET STRINGconfig : config statement\n| statementstatement : table\n| key_value_pair\n| INDENT\n| DEDENTtable : LBRACKET IDENTIFIER RBRACKET\n| LBRACKET IDENTIFIER RBRACKET INDENT statements DEDENTkey_value_pair : IDENTIFIER EQUALS valuevalue : STRING\n| NUMBER\n| DYNAMIC_PLACEHOLDERstatements : statements statement\n| statementindent_statement : INDENTdedent_statement : DEDENT'
    
_lr_action_items = {'INDENT':([0,1,2,3,4,5,6,9,12,13,14,15,16,17,18,19,20,21,],[5,5,-2,-3,-4,-5,-6,-1,17,-9,-10,-11,-12,5,5,-14,-6,-13,]),'DEDENT':([0,1,2,3,4,5,6,9,12,13,14,15,16,17,18,19,20,21,],[6,6,-2,-3,-4,-5,-6,-1,-7,-9,-10,-11,-12,6,20,-14,-6,-13,]),'LBRACKET':([0,1,2,3,4,5,6,9,12,13,14,15,16,17,18,19,20,21,],[7,7,-2,-3,-4,-5,-6,-1,-7,-9,-10,-11,-12,7,7,-14,-6,-13,]),'IDENTIFIER':([0,1,2,3,4,5,6,7,9,12,13,14,15,16,17,18,19,20,21,],[8,8,-2,-3,-4,-5,-6,10,-1,-7,-9,-10,-11,-12,8,8,-14,-6,-13,]),'$end':([1,2,3,4,5,6,9,12,13,14,15,16,20,],[0,-2,-3,-4,-5,-6,-1,-7,-9,-10,-11,-12,-8,]),'EQUALS':([8,],[11,]),'RBRACKET':([10,],[12,]),'STRING':([11,],[14,]),'NUMBER':([11,],[15,]),'DYNAMIC_PLACEHOLDER':([11,],[16,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'config':([0,],[1,]),'statement':([0,1,17,18,],[2,9,19,21,]),'table':([0,1,17,18,],[3,3,3,3,]),'key_value_pair':([0,1,17,18,],[4,4,4,4,]),'value':([11,],[13,]),'statements':([17,],[18,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> config","S'",1,None,None,None),
  ('config -> config statement','config',2,'p_config','_parser.py',98),
  ('config -> statement','config',1,'p_config','_parser.py',99),
  ('statement -> table','statement',1,'p_statement','_parser.py',107),
  ('statement -> key_value_pair','statement',1,'p_statement','_parser.py',108),
  ('statement -> INDENT','statement',1,'p_statement','_parser.py',109),
  ('statement -> DEDENT','statement',1,'p_statement','_parser.py',110),
  ('table -> LBRACKET IDENTIFIER RBRACKET','table',3,'p_table','_parser.py',115),
  ('table -> LBRACKET IDENTIFIER RBRACKET INDENT statements DEDENT','table',6,'p_table','_parser.py',116),
  ('key_value_pair -> IDENTIFIER EQUALS value','key_value_pair',3,'p_key_value_pair','_parser.py',124),
  ('value -> STRING','value',1,'p_value','_parser.py',129),
  ('value -> NUMBER','value',1,'p_value','_parser.py',130),
  ('value -> DYNAMIC_PLACEHOLDER','value',1,'p_value','_parser.py',131),
  ('statements -> statements statement','statements',2,'p_statements','_parser.py',136),
  ('statements -> statement','statements',1,'p_statements','_parser.py',137),
  ('indent_statement -> INDENT','indent_statement',1,'p_indent_statement','_parser.py',146),
  ('dedent_statement -> DEDENT','dedent_statement',1,'p_dedent_statement','_parser.py',151),
]



from collections import deque
from inspect import stack as call_stack

from ..type_func.List import hasindex


class _TreeHead:
    children = []


stack = deque([_TreeHead()])  # type: deque[_TreeHead|CodeStruct]
namespace = {}


def get_called_func(depth=1):
    call_stck = call_stack()
    if not hasindex(call_stck, depth):
        return None
    return call_stck[depth].frame.f_code.co_qualname


class CodeStructMain:
    def __init__(self, name=None, description=None):
        self.name = name
        self.description = description
        self.root = None
        self._register()

    def _register(self):
        if self.name in namespace:
            raise ValueError(f"Duplicate name: {self.name}")
        # Python called stack: _register, __init__, <function>
        funct = get_called_func(3)
        if funct in namespace:
            raise ValueError(f"Duplicate function: {funct}")
        namespace[funct] = self

    def _generate(self, node, depth):
        if not isinstance(node, _TreeHead):
            prefix = '\t' * depth
            res = f"{prefix}{node.subject}: {node.description}\n"
        else:
            res = ""
        for c in node.children:
            res += self._generate(c, depth + 1)

        return res

    def generate_tree_str(self):
        tree_head = self.root.parent
        return self._generate(tree_head, -1)


class CodeStruct:
    def __init__(self, subject=None, description=None):
        self.subject = subject
        self.description = description
        self.children = []
        self.parent = None

        self._called_func = get_called_func(2)
        if self._called_func in namespace:
            namespace.get(self._called_func).root = self
        else:
            raise RuntimeError(f"No root found for {self._called_func}")

    def __enter__(self):
        stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        stack.pop()
        if stack:
            self.parent = stack[-1]
            self.parent.children.append(self)



import os
import shutil
from pathlib import Path
from subprocess import call


def main():
    cwd = Path(__file__).parent
    os.chdir(cwd)

    print('Building...')
    call(['python', 'setup.py', 'build'])

    build_dir = cwd / 'build'
    # lib_dir = list(build_dir.glob('lib*'))[0]
    # ls = list(lib_dir.glob('hybitmap*.pyd'))

    ls = list(build_dir.glob('**/hybitmap*.pyd'))
    file = ls[0] if len(ls) == 1 else None  # type: Path

    if file is None:
        print('No file found')
        return

    print('Copy file...')
    pyd_file_name = (cwd / 'pyd' / file.name)
    pyd_file_name.unlink(True)
    try:
        shutil.copy2(file, pyd_file_name)
    except (FileNotFoundError, PermissionError, OSError) as e:
        print(f'Failed to copy file({e})')
        return
    except Exception as e:
        print('Failed to copy file({e)')
        return

    print('Clean up...')
    shutil.rmtree(build_dir)


if __name__ == '__main__':
    main()



DATABASE_NAME = "SzQlib.db"
ENTRYPT_KEY = B"\x00\x02\xf3\x94\x98\x03\x99\xff\x00\x02\xf3\x94\x98\x03\x99\xff"
STRUCT_SAMPLE = {
    "name": "SzQlib",
    "version": "1.0.0",
    "author": "SzQ",
    "description": "A library for Python",
    "license": "MIT",
    "url": "",
    "email": "",
    "keywords": ["SzQlib", "SzQ", "Python"],
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
}
